@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C)2006-2009 Edgewall Software
+# Copyright (C) 2006-2018 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -125,10 +125,10 @@ class AdminCommandManager(Component):
                     fargs = args[len(parts):]
                     try:
                         return f(*fargs)
-                    except AdminCommandError, e:
+                    except AdminCommandError as e:
                         e.cmd = ' '.join(parts)
                         raise
-                    except TypeError, e:
+                    except TypeError as e:
                         tb = traceback.extract_tb(sys.exc_info()[2])
                         if len(tb) == 1:
                             raise AdminCommandError(_("Invalid arguments"),
@@ -197,19 +197,33 @@ def get_dir_list(path, dirs_only=False):
     return result
 
 
-def get_console_locale(env=None, lang=LANG):
-    """Return negotiated locale for console by LANG environment and
+def get_console_locale(env=None, lang=None,
+                       categories=('LANGUAGE', 'LC_ALL', 'LC_MESSAGES',
+                                   'LANG')):
+    """Return negotiated locale for console by locale environments and
     [trac] default_language."""
     if has_babel:
-        from babel.core import Locale, UnknownLocaleError, parse_locale
+        from babel.core import UnknownLocaleError, parse_locale
+        def normalize(value):
+            if not value:
+                return None
+            try:
+                return '_'.join(filter(None, parse_locale(value)))
+            except:
+                return None
+        locales = [lang]
+        for category in categories:
+            value = os.environ.get(category)
+            if not value:
+                continue
+            if category == 'LANGUAGE' and ':' in value:
+                value = value.split(':')[0]
+            locales.append(value)
+        if env:
+            locales.append(env.config.get('trac', 'default_language'))
+        locales = filter(None, map(normalize, locales))
         try:
-            lang = '_'.join(filter(None, parse_locale(lang)))
-        except:
-            lang = None
-        default = env.config.get('trac', 'default_language', '') \
-                  if env else None
-        try:
-            return get_negotiated_locale([lang, default]) or Locale.default()
+            return get_negotiated_locale(locales)
         except UnknownLocaleError:
             pass
     return None
