@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2009 Edgewall Software
+# Copyright (C) 2005-2018 Edgewall Software
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005 Jeff Weiss <trac@jeffweiss.org>
 # Copyright (C) 2006 Andres Salomon <dilinger@athenacr.com>
@@ -23,10 +23,10 @@ from genshi.core import Markup
 
 from trac.core import *
 from trac.config import Option
-from trac.db.api import DatabaseManager, IDatabaseConnector, _parse_db_str, \
-                        get_column_names
+from trac.db.api import ConnectionBase, DatabaseManager, IDatabaseConnector, \
+                        get_column_names, parse_connection_uri
 from trac.db.util import ConnectionWrapper, IterableCursor
-from trac.env import IEnvironmentSetupParticipant
+from trac.env import IEnvironmentSetupParticipant, ISystemInfoProvider
 from trac.util import as_int, get_pkginfo
 from trac.util.compat import close_fds
 from trac.util.text import exception_to_unicode, to_unicode
@@ -37,39 +37,47 @@ _like_escape_re = re.compile(r'([/_%])')
 try:
     import MySQLdb
     import MySQLdb.cursors
-    has_mysqldb = True
+except ImportError:
+    MySQLdb = None
+    mysqldb_version = None
+else:
+    mysqldb_version = get_pkginfo(MySQLdb).get('version', MySQLdb.__version__)
 
     class MySQLUnicodeCursor(MySQLdb.cursors.Cursor):
         def _convert_row(self, row):
             return tuple(v.decode('utf-8') if isinstance(v, str) else v
                          for v in row)
+
         def fetchone(self):
             row = super(MySQLUnicodeCursor, self).fetchone()
             return self._convert_row(row) if row else None
+
         def fetchmany(self, num):
             rows = super(MySQLUnicodeCursor, self).fetchmany(num)
             return [self._convert_row(row) for row in rows] \
                    if rows is not None else []
+
         def fetchall(self):
             rows = super(MySQLUnicodeCursor, self).fetchall()
             return [self._convert_row(row) for row in rows] \
                    if rows is not None else []
-except ImportError:
-    has_mysqldb = False
 
 # Mapping from "abstract" SQL types to DB-specific types
 _type_map = {
     'int64': 'bigint',
+    'text': 'mediumtext',
 }
 
 
 class MySQLConnector(Component):
     """Database connector for MySQL version 4.1 and greater.
 
-    Database URLs should be of the form:
-    {{{
-    mysql://user[:password]@host[:port]/database[?param1=value&param2=value]
-    }}}
+    Database URLs should be of the form::
+
+      {{{
+      mysql://user[:password]@host[:port]/database[?param1=value&param2=value]
+      }}}
+
     The following parameters are supported:
      * `compress`: Enable compression (0 or 1)
      * `init_command`: Command to run once the connection is created
@@ -78,32 +86,37 @@ class MySQLConnector(Component):
      * `read_default_group`: Configuration group to use from the default file
      * `unix_socket`: Use a Unix socket at the given path to connect
     """
-    implements(IDatabaseConnector, IEnvironmentSetupParticipant)
+    implements(IDatabaseConnector, IEnvironmentSetupParticipant,
+               ISystemInfoProvider)
+
+    required = False
 
     mysqldump_path = Option('trac', 'mysqldump_path', 'mysqldump',
         """Location of mysqldump for MySQL database backups""")
 
     def __init__(self):
-        self._version = None
-        self.error = None
+        self._mysql_version = None
+
+    # ISystemInfoProvider methods
+
+    def get_system_info(self):
+        if self.required:
+            yield 'MySQL', self._mysql_version
+            yield 'MySQLdb', mysqldb_version
+
+    # IDatabaseConnector methods
 
     def get_supported_schemes(self):
-        if not has_mysqldb:
-            self.error = _("Cannot load Python bindings for MySQL")
-        yield ('mysql', -1 if self.error else 1)
+        yield 'mysql', 1
 
     def get_connection(self, path, log=None, user=None, password=None,
                        host=None, port=None, params={}):
         cnx = MySQLConnection(path, log, user, password, host, port, params)
-        if not self._version:
-            self._version = get_pkginfo(MySQLdb).get('version',
-                                                     MySQLdb.__version__)
-            mysql_info = 'server: "%s", client: "%s", thread-safe: %s' % \
-                         (cnx.cnx.get_server_info(),
-                          MySQLdb.get_client_info(),
-                          MySQLdb.thread_safe())
-            self.env.systeminfo.extend([('MySQL', mysql_info),
-                                        ('MySQLdb', self._version)])
+        if not self.required:
+            self._mysql_version = \
+                'server: "%s", client: "%s", thread-safe: %s' \
+                % (cnx.cnx.get_server_info(), MySQLdb.get_client_info(),
+                   MySQLdb.thread_safe())
             self.required = True
         return cnx
 
@@ -115,18 +128,32 @@ class MySQLConnector(Component):
         cnx = self.get_connection(path, log, user, password, host, port,
                                   params)
         self._verify_variables(cnx)
-        utf8_size = self._utf8_size(cnx)
+        max_bytes = self._max_bytes(cnx)
         cursor = cnx.cursor()
         if schema is None:
             from trac.db_default import schema
         for table in schema:
-            for stmt in self.to_sql(table, utf8_size=utf8_size):
+            for stmt in self.to_sql(table, max_bytes=max_bytes):
                 self.log.debug(stmt)
                 cursor.execute(stmt)
         self._verify_table_status(cnx)
         cnx.commit()
 
-    def _utf8_size(self, cnx):
+    def destroy_db(self, path, log=None, user=None, password=None, host=None,
+                   port=None, params={}):
+        cnx = self.get_connection(path, log, user, password, host, port,
+                                  params)
+        for table_name in cnx.get_table_names():
+            cnx.drop_table(table_name)
+        cnx.commit()
+
+    def db_exists(self, path, log=None, user=None, password=None, host=None,
+                  port=None, params={}):
+        cnx = self.get_connection(path, log, user, password, host, port,
+                                  params)
+        return bool(cnx.get_table_names())
+
+    def _max_bytes(self, cnx):
         if cnx is None:
             connector, args = DatabaseManager(self.env).get_connector()
             cnx = connector.get_connection(**args)
@@ -136,28 +163,25 @@ class MySQLConnector(Component):
             charset = cnx.charset
         return 4 if charset == 'utf8mb4' else 3
 
-    def _collist(self, table, columns, utf8_size):
+    _max_key_length = 3072
+
+    def _collist(self, table, columns, max_bytes):
         """Take a list of columns and impose limits on each so that indexing
         works properly.
 
-        Some Versions of MySQL limit each index prefix to 1000 bytes total,
+        Some Versions of MySQL limit each index prefix to 3072 bytes total,
         with a max of 767 bytes per column.
         """
         cols = []
-        limit_col = 767 / utf8_size
-        limit = min(1000 / (utf8_size * len(columns)), limit_col)
+        limit_col = 767 / max_bytes
+        limit = min(self._max_key_length / (max_bytes * len(columns)),
+                    limit_col)
         for c in columns:
             name = '`%s`' % c
             table_col = filter((lambda x: x.name == c), table.columns)
             if len(table_col) == 1 and table_col[0].type.lower() == 'text':
                 if table_col[0].key_size is not None:
                     name += '(%d)' % min(table_col[0].key_size, limit_col)
-                elif name == '`rev`':
-                    name += '(20)'
-                elif name == '`path`':
-                    name += '(%d)' % limit_col
-                elif name == '`change_type`':
-                    name += '(2)'
                 else:
                     name += '(%s)' % limit
             # For non-text columns, we simply throw away the extra bytes.
@@ -165,9 +189,9 @@ class MySQLConnector(Component):
             cols.append(name)
         return ','.join(cols)
 
-    def to_sql(self, table, utf8_size=None):
-        if utf8_size is None:
-            utf8_size = self._utf8_size(None)
+    def to_sql(self, table, max_bytes=None):
+        if max_bytes is None:
+            max_bytes = self._max_bytes(None)
         sql = ['CREATE TABLE %s (' % table.name]
         coldefs = []
         for column in table.columns:
@@ -182,15 +206,15 @@ class MySQLConnector(Component):
         if len(table.key) > 0:
             coldefs.append('    PRIMARY KEY (%s)' %
                            self._collist(table, table.key,
-                                         utf8_size=utf8_size))
+                                         max_bytes=max_bytes))
         sql.append(',\n'.join(coldefs) + '\n)')
         yield '\n'.join(sql)
 
         for index in table.indices:
             unique = 'UNIQUE' if index.unique else ''
-            yield 'CREATE %s INDEX %s_%s_idx ON %s (%s);' % (unique, table.name,
+            yield 'CREATE %s INDEX %s_%s_idx ON %s (%s)' % (unique, table.name,
                   '_'.join(index.columns), table.name,
-                  self._collist(table, index.columns, utf8_size=utf8_size))
+                  self._collist(table, index.columns, max_bytes=max_bytes))
 
     def alter_column_types(self, table, columns):
         """Yield SQL statements altering the type of one or more columns of
@@ -212,11 +236,11 @@ class MySQLConnector(Component):
     def backup(self, dest_file):
         from subprocess import Popen, PIPE
         db_url = self.env.config.get('trac', 'database')
-        scheme, db_prop = _parse_db_str(db_url)
+        scheme, db_prop = parse_connection_uri(db_url)
         db_params = db_prop.setdefault('params', {})
         db_name = os.path.basename(db_prop['path'])
 
-        args = [self.mysqldump_path]
+        args = [self.mysqldump_path, '--no-defaults']
         if 'host' in db_prop:
             args.extend(['-h', db_prop['host']])
         if 'port' in db_prop:
@@ -228,7 +252,7 @@ class MySQLConnector(Component):
                 args.append('--compress')
             elif name == 'named_pipe' and as_int(value, 0):
                 args.append('--protocol=pipe')
-            elif name == 'read_default_file': # Must be first
+            elif name == 'read_default_file':  # Must be first
                 args.insert(1, '--defaults-file=' + value)
             elif name == 'unix_socket':
                 args.extend(['--protocol=socket', '--socket=' + value])
@@ -242,7 +266,7 @@ class MySQLConnector(Component):
             environ['MYSQL_PWD'] = str(db_prop['password'])
         try:
             p = Popen(args, env=environ, stderr=PIPE, close_fds=close_fds)
-        except OSError, e:
+        except OSError as e:
             raise TracError(_("Unable to run %(path)s: %(msg)s",
                               path=self.mysqldump_path,
                               msg=exception_to_unicode(e)))
@@ -259,13 +283,14 @@ class MySQLConnector(Component):
     def environment_created(self):
         pass
 
-    def environment_needs_upgrade(self, db):
-        if getattr(self, 'required', False):
-            self._verify_table_status(db)
-            self._verify_variables(db)
+    def environment_needs_upgrade(self):
+        if self.required:
+            with self.env.db_query as db:
+                self._verify_table_status(db)
+                self._verify_variables(db)
         return False
 
-    def upgrade_environment(self, db):
+    def upgrade_environment(self):
         pass
 
     UNSUPPORTED_ENGINES = ('MyISAM', 'EXAMPLE', 'ARCHIVE', 'CSV', 'ISAM')
@@ -332,7 +357,7 @@ class MySQLConnector(Component):
                 supported=repr(self.SUPPORTED_COLLATIONS)))
 
 
-class MySQLConnection(ConnectionWrapper):
+class MySQLConnection(ConnectionBase, ConnectionWrapper):
     """Connection wrapper for MySQL."""
 
     poolable = True
@@ -341,9 +366,9 @@ class MySQLConnection(ConnectionWrapper):
                  port=None, params={}):
         if path.startswith('/'):
             path = path[1:]
-        if password == None:
+        if password is None:
             password = ''
-        if port == None:
+        if port is None:
             port = 3306
         opts = {}
         for name, value in params.iteritems():
@@ -394,7 +419,7 @@ class MySQLConnection(ConnectionWrapper):
             self._is_closed = True
 
     def cast(self, column, type):
-        if type == 'int' or type == 'int64':
+        if type in ('int', 'int64'):
             type = 'signed'
         elif type == 'text':
             type = 'char'
@@ -403,6 +428,28 @@ class MySQLConnection(ConnectionWrapper):
     def concat(self, *args):
         return 'concat(%s)' % ', '.join(args)
 
+    def drop_column(self, table, column):
+        cursor = MySQLdb.cursors.Cursor(self.cnx)
+        if column in self.get_column_names(table):
+            quoted_table = self.quote(table)
+            cursor.execute("SHOW INDEX FROM %s" % quoted_table)
+            columns = get_column_names(cursor)
+            keys = {}
+            for row in cursor.fetchall():
+                row = dict(zip(columns, row))
+                keys.setdefault(row['Key_name'], []).append(row['Column_name'])
+            # drop all composite indices which in the given column is involved
+            for key, columns in keys.iteritems():
+                if len(columns) > 1 and column in columns:
+                    if key == 'PRIMARY':
+                        cursor.execute("ALTER TABLE %s DROP PRIMARY KEY" %
+                                       quoted_table)
+                    else:
+                        cursor.execute("ALTER TABLE %s DROP KEY %s" %
+                                       (quoted_table, self.quote(key)))
+            cursor.execute("ALTER TABLE %s DROP COLUMN %s " %
+                           (quoted_table, self.quote(column)))
+
     def drop_table(self, table):
         cursor = MySQLdb.cursors.Cursor(self.cnx)
         cursor._defer_warnings = True  # ignore "Warning: Unknown table ..."
@@ -412,6 +459,7 @@ class MySQLConnection(ConnectionWrapper):
         rows = self.execute("""
             SELECT column_name FROM information_schema.columns
             WHERE table_schema=%s AND table_name=%s
+            ORDER BY ordinal_position
             """, (self.schema, table))
         return [row[0] for row in rows]
 
@@ -421,26 +469,44 @@ class MySQLConnection(ConnectionWrapper):
     def get_table_names(self):
         rows = self.execute("""
             SELECT table_name FROM information_schema.tables
-            WHERE table_schema=%s""", (self.schema,))
+            WHERE table_schema=%s
+            """, (self.schema,))
         return [row[0] for row in rows]
 
     def like(self):
-        """Return a case-insensitive LIKE clause."""
         return "LIKE %%s COLLATE %s_general_ci ESCAPE '/'" % self.charset
 
     def like_escape(self, text):
         return _like_escape_re.sub(r'/\1', text)
 
+    def reset_tables(self):
+        table_names = []
+        if not self.schema:
+            return table_names
+        cursor = self.cursor()
+        cursor.execute("""
+            SELECT table_name, auto_increment
+            FROM information_schema.tables
+            WHERE table_schema=%s""", (self.schema,))
+        for table, auto_increment in cursor.fetchall():
+            table_names.append(table)
+            if auto_increment is None or auto_increment == 1:
+                # DELETE FROM is preferred to TRUNCATE TABLE, as the
+                # auto_increment is not used or it is 1.
+                cursor.execute("DELETE FROM %s" % table)
+            else:
+                # TRUNCATE TABLE is preferred to DELETE FROM, as we
+                # need to reset the auto_increment in MySQL.
+                cursor.execute("TRUNCATE TABLE %s" % table)
+        return table_names
+
     def prefix_match(self):
-        """Return a case sensitive prefix-matching operator."""
         return "LIKE %s ESCAPE '/'"
 
     def prefix_match_value(self, prefix):
-        """Return a value for case sensitive prefix-matching operator."""
         return self.like_escape(prefix) + '%'
 
     def quote(self, identifier):
-        """Return the quoted identifier."""
         return "`%s`" % identifier.replace('`', '``')
 
     def update_sequence(self, cursor, table, column='id'):
