@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C)2005-2009 Edgewall Software
+# Copyright (C) 2005-2018 Edgewall Software
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -14,19 +14,18 @@
 #
 # Author: Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import with_statement
-
 import os
-import time
+import sys
 
 from trac.core import TracError
 from trac.db.util import ConnectionWrapper
 from trac.util.concurrency import threading
+from trac.util.datefmt import time_now
 from trac.util.text import exception_to_unicode
 from trac.util.translation import _
 
 
-class TimeoutError(Exception):
+class TimeoutError(TracError):
     """Exception raised by the connection pool when no connection has become
     available after a given timeout."""
 
@@ -70,7 +69,7 @@ class ConnectionPoolBackend(object):
         cnx = None
         log = kwargs.get('log')
         key = unicode(kwargs)
-        start = time.time()
+        start = time_now()
         tid = threading._get_ident()
         # Get a Connection, either directly or a deferred one
         with self._available:
@@ -91,9 +90,9 @@ class ConnectionPoolBackend(object):
                 self._active[(tid, key)] = (cnx, num)
 
         deferred = num == 1 and isinstance(cnx, tuple)
-        err = None
+        exc_info = (None, None, None)
         if deferred:
-            # Potentially lenghty operations must be done without lock held
+            # Potentially lengthy operations must be done without lock held
             op, cnx = cnx
             try:
                 if op == 'ping':
@@ -102,16 +101,16 @@ class ConnectionPoolBackend(object):
                     cnx.close()
                 if op in ('close', 'create'):
                     cnx = connector.get_connection(**kwargs)
-            except TracError, e:
-                err = e
+            except TracError:
+                exc_info = sys.exc_info()
                 cnx = None
-            except Exception, e:
+            except Exception:
+                exc_info = sys.exc_info()
                 if log:
                     log.error('Exception caught on %s', op, exc_info=True)
-                err = e
                 cnx = None
 
-        if cnx:
+        if cnx and not isinstance(cnx, tuple):
             if deferred:
                 # replace placeholder with real Connection
                 with self._available:
@@ -126,11 +125,13 @@ class ConnectionPoolBackend(object):
                 return self.get_cnx(connector, kwargs)
 
         # if we didn't get a cnx after wait(), something's fishy...
-        timeout = time.time() - start
+        if isinstance(exc_info[1], TracError):
+            raise exc_info[0], exc_info[1], exc_info[2]
+        timeout = time_now() - start
         errmsg = _("Unable to get database connection within %(time)d seconds.",
                    time=timeout)
-        if err:
-            errmsg += " (%s)" % exception_to_unicode(err)
+        if exc_info[1]:
+            errmsg += " (%s)" % exception_to_unicode(exc_info[1])
         raise TimeoutError(errmsg)
 
     def _take_cnx(self, connector, kwargs, key, tid):
@@ -178,7 +179,7 @@ class ConnectionPoolBackend(object):
                 if cnx and cnx.poolable:
                     self._pool.append(cnx)
                     self._pool_key.append(key)
-                    self._pool_time.append(time.time())
+                    self._pool_time.append(time_now())
                 self._available.notify()
 
     def shutdown(self, tid=None):
@@ -186,7 +187,7 @@ class ConnectionPoolBackend(object):
         delay = 120
         if tid is None:
             delay = 0
-        when = time.time() - delay
+        when = time_now() - delay
         with self._available:
             if tid is None: # global shutdown, also close active connections
                 for db, num in self._active.values():
@@ -214,4 +215,3 @@ class ConnectionPool(object):
 
     def shutdown(self, tid=None):
         _backend.shutdown(tid)
-
