@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C)2005-2009 Edgewall Software
+# Copyright (C) 2005-2018 Edgewall Software
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -22,24 +22,34 @@ class Table(object):
         self.name = name
         self.columns = []
         self.indices = []
-        self.key = key
-        if isinstance(key, basestring):
-            self.key = [key]
+        self.key = [key] if isinstance(key, basestring) else key
 
     def __getitem__(self, objs):
         self.columns = [o for o in objs if isinstance(o, Column)]
         self.indices = [o for o in objs if isinstance(o, Index)]
         return self
 
+    def remove_columns(self, column_names):
+        """Remove columns specified in the list or tuple `column_names`."""
+        if not isinstance(column_names, (list, tuple)):
+            column_names = [column_names]
+        if any(c in column_names for c in self.key):
+            self.key = []
+        self.columns = [col for col in self.columns
+                        if col.name not in column_names]
+        self.indices = [idx for idx in self.indices
+                        if all(c not in column_names for c in idx.columns)]
+
 
 class Column(object):
     """Declare a table column in a database schema."""
 
-    def __init__(self, name, type='text', size=None, unique=False,
+    def __init__(self, name, type='text', size=None, key_size=None,
                  auto_increment=False):
         self.name = name
         self.type = type
         self.size = size
+        self.key_size = key_size
         self.auto_increment = auto_increment
 
 
