@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2014 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2007 Jonas Borgström <jonas@edgewall.com>
 # All rights reserved.
 #
@@ -27,7 +27,7 @@ from trac import db_default, log
 from trac.admin import AdminCommandError, IAdminCommandProvider
 from trac.cache import CacheManager, cached
 from trac.config import BoolOption, ChoiceOption, ConfigSection, \
-                        Configuration, Option, PathOption
+                        Configuration, IntOption, Option, PathOption
 from trac.core import Component, ComponentManager, implements, Interface, \
                       ExtensionPoint, TracBaseError, TracError
 from trac.db.api import (DatabaseManager, QueryContextManager,
@@ -35,7 +35,7 @@ from trac.db.api import (DatabaseManager, QueryContextManager,
 from trac.loader import load_components
 from trac.log import logger_handler_factory
 from trac.util import arity, as_bool, copytree, create_file, get_pkginfo, \
-                      lazy, makedirs, read_file
+                      is_path_below, lazy, makedirs, read_file
 from trac.util.concurrency import threading
 from trac.util.text import exception_to_unicode, path_to_unicode, printerr, \
                            printout
@@ -170,6 +170,9 @@ class Environment(Component, ComponentManager):
         Plugins in that directory are loaded in addition to those in
         the directory of the environment `plugins`, with this one
         taking precedence.
+
+        Non-absolute paths are relative to the Environment `conf`
+        directory.
         """)
 
     base_url = Option('trac', 'base_url', '',
@@ -201,6 +204,13 @@ class Environment(Component, ComponentManager):
         your Trac instance is only accessible through HTTPS.
         """)
 
+    anonymous_session_lifetime = IntOption(
+        'trac', 'anonymous_session_lifetime', '90',
+        """Lifetime of the anonymous session, in days.
+
+        Set the option to 0 to disable purging old anonymous sessions.
+        (''since 1.0.17'')""")
+
     project_name = Option('project', 'name', 'My Project',
         """Name of the project.""")
 
@@ -237,7 +247,8 @@ class Environment(Component, ComponentManager):
                             log.LOG_TYPES + log.LOG_TYPE_ALIASES,
         """Logging facility to use.
 
-        Should be one of (`none`, `file`, `stderr`, `syslog`, `winlog`).""")
+        Should be one of (`none`, `file`, `stderr`, `syslog`, `winlog`).""",
+        case_sensitive=False)
 
     log_file = Option('logging', 'log_file', 'trac.log',
         """If `log_type` is `file`, this should be a path to the
@@ -250,7 +261,7 @@ class Environment(Component, ComponentManager):
         """Level of verbosity in log.
 
         Should be one of (`CRITICAL`, `ERROR`, `WARNING`, `INFO`, `DEBUG`).
-        """)
+        """, case_sensitive=False)
 
     log_format = Option('logging', 'log_format', None,
         """Custom logging format.
@@ -727,18 +738,23 @@ class Environment(Component, ComponentManager):
         logfile = self.log_file
         if logtype == 'file' and not os.path.isabs(logfile):
             logfile = os.path.join(self.log_dir, logfile)
-        format = self.log_format
-        logid = 'Trac.%s' % hashlib.sha1(self.path).hexdigest()
-        if format:
-            format = format.replace('$(', '%(') \
-                           .replace('%(path)s', self.path) \
-                           .replace('%(basename)s', self.name) \
-                           .replace('%(project)s', self.project_name)
-        self.log, self._log_handler = logger_handler_factory(
-            logtype, logfile, self.log_level, logid, format=format)
+        self.log, self._log_handler = \
+            self.create_logger(self.log_type, logfile, self.log_level,
+                               self.log_format)
+        self.log.addHandler(self._log_handler)
         self.log.info('-' * 32 + ' environment startup [Trac %s] ' + '-' * 32,
                       self.trac_version)
 
+    def create_logger(self, log_type, log_file, log_level, log_format):
+        log_id = 'Trac.%s' % hashlib.sha1(self.path).hexdigest()
+        if log_format:
+            log_format = log_format.replace('$(', '%(') \
+                                   .replace('%(path)s', self.path) \
+                                   .replace('%(basename)s', self.name) \
+                                   .replace('%(project)s', self.project_name)
+        return logger_handler_factory(log_type, log_file, log_level, log_id,
+                                      format=log_format)
+
     def get_known_users(self, as_dict=False):
         """Returns information about all known users, i.e. users that
         have logged in to this Trac environment and possibly set their
@@ -829,7 +845,9 @@ class Environment(Component, ComponentManager):
                     args = (db,)
                 participant.upgrade_environment(*args)
             # Database schema may have changed, so close all connections
-            DatabaseManager(self).shutdown()
+            dbm = DatabaseManager(self)
+            if dbm.connection_uri != 'sqlite::memory:':
+                dbm.shutdown()
         del self.database_version
         return True
 
@@ -865,8 +883,7 @@ class EnvironmentSetup(Component):
         return DatabaseManager(self.env).needs_upgrade(db_default.db_version)
 
     def upgrade_environment(self):
-        DatabaseManager(self.env).upgrade(db_default.db_version,
-                                          pkg='trac.upgrades')
+        DatabaseManager(self.env).upgrade(db_default.db_version)
         self._update_sample_config()
 
     # Internal methods
@@ -876,14 +893,13 @@ class EnvironmentSetup(Component):
         if not os.path.isfile(filename):
             return
         config = Configuration(filename)
-        for (section, name), option in Option.get_registry().iteritems():
-            config.set(section, name, option.dumps(option.default))
+        config.set_defaults()
         try:
             config.save()
             self.log.info("Wrote sample configuration file with the new "
                           "settings and their default values: %s",
                           filename)
-        except IOError as e:
+        except EnvironmentError as e:
             self.log.warn("Couldn't write sample configuration file (%s)", e,
                           exc_info=True)
 
@@ -936,8 +952,8 @@ def open_environment(env_path=None, use_cache=False):
             env.log.error("Exception caught while checking for upgrade: %s",
                           exception_to_unicode(e, traceback=True))
         if needs_upgrade:
-            raise TracError(_('The Trac Environment needs to be upgraded.\n\n'
-                              'Run "trac-admin %(path)s upgrade"',
+            raise TracError(_('The Trac Environment needs to be upgraded. '
+                              'Run:\n\n  trac-admin "%(path)s" upgrade',
                               path=env_path))
 
     return env
@@ -965,25 +981,45 @@ class EnvironmentAdmin(Component):
         yield ('upgrade', '[--no-backup]',
                """Upgrade database to current version
 
-               The database is backed up to the environment 'db' directory,
-               unless the --no-backup option is specified. The shorthand alias
-               -b can also be used to specify --no-backup.
+               The database is backed up to the directory specified by [trac]
+               backup_dir (the default is 'db'), unless the --no-backup
+               option is specified. The shorthand alias -b can also be used
+               to specify --no-backup.
                """,
                None, self._do_upgrade)
 
     def _do_deploy(self, dest):
+        from trac.web.chrome import Chrome
+
         target = os.path.normpath(dest)
         chrome_target = os.path.join(target, 'htdocs')
         script_target = os.path.join(target, 'cgi-bin')
 
+        # Check source and destination to avoid recursively copying files
+        for provider in Chrome(self.env).template_providers:
+            paths = list(provider.get_htdocs_dirs() or [])
+            if not paths:
+                continue
+            for key, root in paths:
+                if not root:
+                    continue
+                source = os.path.normpath(root)
+                dest = os.path.join(chrome_target, key)
+                if os.path.exists(source) and is_path_below(dest, source):
+                    raise AdminCommandError(
+                        _("Resources cannot be deployed to a target "
+                          "directory that is equal to or below the source "
+                          "directory '%(source)s'.\n\nPlease choose a "
+                          "different target directory and try again.",
+                          source=source))
+
         # Copy static content
         makedirs(target, overwrite=True)
         makedirs(chrome_target, overwrite=True)
-        from trac.web.chrome import Chrome
         printout(_("Copying resources from:"))
         for provider in Chrome(self.env).template_providers:
             paths = list(provider.get_htdocs_dirs() or [])
-            if not len(paths):
+            if not paths:
                 continue
             printout('  %s.%s' % (provider.__module__,
                                   provider.__class__.__name__))
@@ -1100,7 +1136,7 @@ class EnvironmentAdmin(Component):
                                "anymore. Please remove it by hand.",
                                err=exception_to_unicode(e)))
 
-        printout(_("Upgrade done.\n\n"
-                   "You may want to upgrade the Trac documentation now by "
-                   "running:\n\n  trac-admin %(path)s wiki upgrade",
+        printout(_('Upgrade done.\n\n'
+                   'You may want to upgrade the Trac documentation now by '
+                   'running:\n\n  trac-admin "%(path)s" wiki upgrade',
                    path=path_to_unicode(self.env.path)))
