@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -60,7 +60,8 @@ def search_to_sql(db, columns, terms):
 
 def search_to_regexps(terms):
     """Convert search query terms into regular expressions."""
-    return [re.compile(re.escape(term)) for term in terms]
+    return [re.compile(re.escape(term), re.IGNORECASE | re.UNICODE)
+            for term in terms]
 
 
 def shorten_result(text='', keywords=[], maxlen=240, fuzz=60):
@@ -70,7 +71,7 @@ def shorten_result(text='', keywords=[], maxlen=240, fuzz=60):
     beg = -1
     for k in keywords:
         i = text_low.find(k.lower())
-        if (i > -1 and i < beg) or beg == -1:
+        if (-1 < i < beg) or beg == -1:
             beg = i
     excerpt_beg = 0
     if beg > fuzz:
@@ -88,5 +89,5 @@ def shorten_result(text='', keywords=[], maxlen=240, fuzz=60):
     if beg > fuzz:
         msg = '... ' + msg
     if beg < len(text)-maxlen:
-        msg = msg + ' ...'
+        msg += ' ...'
     return msg
