@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # All rights reserved.
 #
@@ -27,7 +27,7 @@ from trac.config import (
 from trac.core import *
 from trac.perm import IPermissionRequestor, PermissionCache, PermissionSystem
 from trac.resource import IResourceManager
-from trac.util import Ranges, as_int
+from trac.util import Ranges, as_bool, as_int
 from trac.util.text import shorten_line
 from trac.util.translation import _, N_, deactivate, gettext, reactivate
 from trac.wiki import IWikiSyntaxProvider, WikiParser
@@ -85,8 +85,7 @@ class ITicketActionController(Interface):
         field this action controller knows about.
 
         This will be used to populate the query options and the like.
-        It is assumed that the initial status of a ticket is 'new' and
-        the terminal status of a ticket is 'closed'.
+        It is assumed that the terminal status of a ticket is 'closed'.
         """
 
     def render_ticket_action_control(req, ticket, action):
@@ -380,21 +379,15 @@ class TicketSystem(Component):
                 self.log.warning('Duplicate field name "%s" (ignoring)',
                                  field['name'])
                 continue
-            if field['name'] in self.reserved_field_names:
-                self.log.warning('Field name "%s" is a reserved name '
-                                 '(ignoring)', field['name'])
-                continue
-            if not re.match('^[a-zA-Z][a-zA-Z0-9_]+$', field['name']):
-                self.log.warning('Invalid name for custom field: "%s" '
-                                 '(ignoring)', field['name'])
-                continue
             fields.append(field)
 
         return fields
 
     reserved_field_names = ['report', 'order', 'desc', 'group', 'groupdesc',
                             'col', 'row', 'format', 'max', 'page', 'verbose',
-                            'comment', 'or']
+                            'comment', 'or', 'id', 'time', 'changetime',
+                            'owner', 'reporter', 'cc', 'summary',
+                            'description', 'keywords']
 
     def get_custom_fields(self):
         return copy.deepcopy(self.custom_fields)
@@ -422,6 +415,8 @@ class TicketSystem(Component):
                     field['optional'] = True
                     if '' in field['options']:
                         field['options'].remove('')
+            elif field['type'] == 'checkbox':
+                field['value'] = '1' if as_bool(field['value']) else '0'
             elif field['type'] == 'text':
                 field['format'] = config.get(name + '.format', 'plain')
             elif field['type'] == 'textarea':
@@ -429,6 +424,16 @@ class TicketSystem(Component):
                 field['height'] = config.getint(name + '.rows')
             elif field['type'] == 'time':
                 field['format'] = config.get(name + '.format', 'datetime')
+
+            if field['name'] in self.reserved_field_names:
+                self.log.warning('Field name "%s" is a reserved name '
+                                 '(ignoring)', field['name'])
+                continue
+            if not re.match('^[a-zA-Z][a-zA-Z0-9_]+$', field['name']):
+                self.log.warning('Invalid name for custom field: "%s" '
+                                 '(ignoring)', field['name'])
+                continue
+
             fields.append(field)
 
         fields.sort(lambda x, y: cmp((x['order'], x['name']),
