@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2005 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2005-2006 Emmanuel Blot <emmanuel.blot@free.fr>
 # All rights reserved.
@@ -16,66 +16,73 @@
 # Author: Daniel Lundin <daniel@edgewall.com>
 #
 
-from __future__ import with_statement
-
-from contextlib import contextmanager
-from hashlib import md5
-
 from genshi.template.text import NewTextTemplate
 
 from trac.attachment import IAttachmentChangeListener
 from trac.core import *
 from trac.config import *
-from trac.notification import NotifyEmail
-from trac.ticket.api import TicketSystem
+from trac.env import IEnvironmentSetupParticipant
+from trac.notification.api import (IEmailDecorator, INotificationFormatter,
+                                   INotificationSubscriber,
+                                   NotificationEvent, NotificationSystem)
+from trac.notification.compat import NotifyEmail
+from trac.notification.mail import (RecipientMatcher, create_message_id,
+                                    get_from_author, set_header)
+from trac.notification.model import Subscription
+from trac.ticket.api import translation_deactivated
 from trac.ticket.model import Ticket
-from trac.util.datefmt import datetime_now, to_utimestamp, utc
+from trac.util.datefmt import (datetime_now, format_date_or_datetime,
+                               get_timezone, utc)
 from trac.util.text import exception_to_unicode, obfuscate_email_address, \
                            shorten_line, text_width, wrap
-from trac.util.translation import deactivate, reactivate
+from trac.util.translation import _
+from trac.web.chrome import Chrome
 
 
 class TicketNotificationSystem(Component):
 
-    always_notify_owner = BoolOption('notification', 'always_notify_owner',
-                                     'false',
-        """Always send notifications to the ticket owner (''since 0.9'').""")
+    implements(IEnvironmentSetupParticipant)
 
-    always_notify_reporter = BoolOption('notification',
-                                        'always_notify_reporter',
-                                        'false',
-        """Always send notifications to any address in the ''reporter''
-        field.""")
+    def environment_created(self):
+        section = 'notification-subscriber'
+        if section not in self.config.sections():
+            self.config.set(section, 'always_notify_cc',
+                            'CarbonCopySubscriber')
+            self.config.set(section, 'always_notify_updater',
+                            'TicketUpdaterSubscriber')
+            self.config.set(section, 'always_notify_previous_updater',
+                            'TicketPreviousUpdatersSubscriber')
+            self.config.save()
 
-    always_notify_updater = BoolOption('notification', 'always_notify_updater',
-                                       'true',
-        """Always send notifications to the person who causes the ticket
-        property change and to any previous updater of that ticket.""")
+    def environment_needs_upgrade(self):
+        return False
+
+    def upgrade_environment(self):
+        pass
 
     ticket_subject_template = Option('notification', 'ticket_subject_template',
-                                     '$prefix #$ticket.id: $summary',
-        """A Genshi text template snippet used to get the notification subject.
+                                     '${prefix} #${ticket.id}: ${summary}',
+        """A Genshi text template snippet used to get the notification
+        subject.
 
-        By default, the subject template is `$prefix #$ticket.id: $summary`.
-        `$prefix` being the value of the `smtp_subject_prefix` option.
-        ''(since 0.11)''""")
+        The template variables are documented on the
+        [TracNotification#Customizingthee-mailsubject TracNotification] page.
+        """)
 
     batch_subject_template = Option('notification', 'batch_subject_template',
-                                    '$prefix Batch modify: $tickets_descr',
-        """Like ticket_subject_template but for batch modifications.
-
-        By default, the template is `$prefix Batch modify: $tickets_descr`.
-        ''(since 1.0)''""")
+                                    '${prefix} Batch modify: ${tickets_descr}',
+        """Like `ticket_subject_template` but for batch modifications.
+        (''since 1.0'')""")
 
     ambiguous_char_width = Option('notification', 'ambiguous_char_width',
                                   'single',
-        """Which width of ambiguous characters (e.g. 'single' or
-        'double') should be used in the table of notification mail.
+        """Width of ambiguous characters that should be used in the table
+        of the notification mail.
 
-        If 'single', the same width as characters in US-ASCII. This is
-        expected by most users. If 'double', twice the width of
-        US-ASCII characters.  This is expected by CJK users. ''(since
-        0.12.2)''""")
+        If `single`, the same width as characters in US-ASCII. This is
+        expected by most users. If `double`, twice the width of
+        US-ASCII characters.  This is expected by CJK users. (''since
+        0.12.2'')""")
 
 
 def get_ticket_notification_recipients(env, config, tktid, prev_cc=None,
@@ -84,6 +91,7 @@ def get_ticket_notification_recipients(env, config, tktid, prev_cc=None,
 
     :since 1.0.2: the `config` parameter is no longer used.
     :since 1.0.2: the `prev_cc` parameter is deprecated.
+    :since 1.1.3: deprecated and will be removed in 1.3.1.
     """
     section = env.config['notification']
     always_notify_reporter = section.getbool('always_notify_reporter')
@@ -143,8 +151,376 @@ def get_ticket_notification_recipients(env, config, tktid, prev_cc=None,
            tkt['reporter'], tkt['owner']
 
 
+class TicketChangeEvent(NotificationEvent):
+    """Represent a ticket change `NotificationEvent`."""
+
+    def __init__(self, category, target, time, author, comment=None,
+                 changes=None, attachment=None):
+        super(TicketChangeEvent, self).__init__('ticket', category, target,
+                                                time, author)
+        self.comment = comment
+        if changes is None and time is not None:
+            changes = target.get_change(cdate=time)
+        self.changes = changes  or {}
+        self.attachment = attachment
+
+
+class BatchTicketChangeEvent(NotificationEvent):
+    """Represent a ticket batch modify `NotificationEvent`."""
+
+    def __init__(self, targets, time, author, comment, new_values, action):
+        super(BatchTicketChangeEvent, self).__init__('ticket', 'batchmodify',
+                                                     targets, time, author)
+        self.comment = comment
+        self.new_values = new_values
+        self.action = action
+
+    def get_ticket_change_events(self, env):
+        for id in self.target:
+            model = Ticket(env, id)
+            yield TicketChangeEvent('changed', model, self.time, self.author,
+                                    self.comment)
+
+
+class TicketFormatter(Component):
+    """Format `TicketChangeEvent` notifications."""
+
+    implements(INotificationFormatter, IEmailDecorator)
+
+    def get_supported_styles(self, transport):
+        yield 'text/plain', 'ticket'
+
+    def format(self, transport, style, event):
+        if event.realm != 'ticket':
+            return
+        if event.category == 'batchmodify':
+            return self._format_plaintext_batchmodify(event)
+        if event.category in ('attachment added', 'attachment deleted'):
+            return self._format_plaintext_attachment(event)
+        else:
+            return self._format_plaintext(event)
+
+    def _format_plaintext(self, event):
+        notify = TicketNotifyEmail(self.env)
+        return notify.format(event.target, event.category == 'created',
+                             event.time)
+
+    def _format_plaintext_attachment(self, event):
+        notify = TicketNotifyEmail(self.env)
+        return notify.format_attachment(event.target, event.attachment,
+                                        event.category == 'attachment added')
+
+    def _format_plaintext_batchmodify(self, event):
+        notify = BatchTicketNotifyEmail(self.env)
+        return notify.format(event.target, event.new_values, event.comment,
+                             event.action, event.author, event.time)
+
+    def _get_message_id(self, event, newticket=None):
+        ticket = event.target
+        from_email = get_from_author(self.env, event)
+        if from_email and isinstance(from_email, tuple):
+            from_email = from_email[1]
+        if not from_email:
+            from_email = self.config.get('notification', 'smtp_from') or \
+                         self.config.get('notification', 'smtp_replyto')
+        modtime = None if newticket else event.time
+        return create_message_id(self.env, '%08d' % ticket.id, from_email,
+                                 modtime, ticket['reporter'] or '')
+
+    def decorate_message(self, event, message, charset):
+        if event.realm != 'ticket':
+            return
+        if event.category == 'batchmodify':
+            notify = BatchTicketNotifyEmail(self.env)
+            tickets = event.target
+            tickets_descr = ', '.join(['#%s' % t for t in tickets])
+            subject = notify.format_subj(tickets_descr)
+        else:
+            notify = TicketNotifyEmail(self.env)
+            ticket = event.target
+            notify.ticket = ticket
+            summary = ticket['summary']
+            from trac.ticket.web_ui import TicketModule
+            for change in TicketModule(self.env) \
+                          .grouped_changelog_entries(ticket, when=event.time):
+                if 'summary' in change['fields']:
+                    values = change['fields']['summary']
+                    summary = "%s (was: %s)" % (values['new'], values['old'])
+            subject = notify.format_subj(summary, event.category == 'created')
+            msgid = self._get_message_id(event, newticket=True)
+            url = self.env.abs_href.ticket(ticket.id)
+            if event.category == 'created':
+                set_header(message, 'Message-ID', msgid, charset)
+            else:
+                set_header(message, 'Message-ID', self._get_message_id(event),
+                           charset)
+                set_header(message, 'In-Reply-To', msgid, charset)
+                set_header(message, 'References', msgid, charset)
+                cnum = ticket.get_comment_number(event.time)
+                if cnum is not None:
+                    url += '#comment:%d' % cnum
+            set_header(message, 'X-Trac-Ticket-ID', ticket.id, charset)
+            set_header(message, 'X-Trac-Ticket-URL', url, charset)
+        set_header(message, 'Subject', subject, charset)
+
+
+class TicketOwnerSubscriber(Component):
+    """Allows ticket owners to subscribe to their tickets."""
+
+    implements(INotificationSubscriber)
+
+    def matches(self, event):
+        if event.realm != 'ticket':
+            return
+        if event.category not in ('created', 'changed', 'attachment added',
+                                  'attachment deleted'):
+            return
+        ticket = event.target
+
+        owners = [ticket['owner']]
+
+        # Harvest previous owner
+        if 'fields' in event.changes and 'owner' in event.changes['fields']:
+            owners.append(event.changes['fields']['owner']['old'])
+
+        matcher = RecipientMatcher(self.env)
+        klass = self.__class__.__name__
+        sids = set()
+        for owner in owners:
+            recipient = matcher.match_recipient(owner)
+            if not recipient:
+                continue
+            sid, auth, addr = recipient
+
+            # Default subscription
+            for s in self.default_subscriptions():
+                yield s[0], s[1], sid, auth, addr, s[2], s[3], s[4]
+            if sid:
+                sids.add((sid,auth))
+
+        for s in Subscription.find_by_sids_and_class(self.env, sids, klass):
+            yield s.subscription_tuple()
+
+    def description(self):
+        return _("Ticket that I own is created or modified")
+
+    def default_subscriptions(self):
+        klass = self.__class__.__name__
+        return NotificationSystem(self.env).default_subscriptions(klass)
+
+    def requires_authentication(self):
+        return True
+
+
+class TicketUpdaterSubscriber(Component):
+    """Allows updaters to subscribe to their own updates."""
+
+    implements(INotificationSubscriber)
+
+    def matches(self, event):
+        if event.realm != 'ticket':
+            return
+        if event.category not in ('created', 'changed', 'attachment added',
+                                  'attachment deleted'):
+            return
+
+        matcher = RecipientMatcher(self.env)
+        recipient = matcher.match_recipient(event.author)
+        if not recipient:
+            return
+        sid, auth, addr = recipient
+
+        # Default subscription
+        for s in self.default_subscriptions():
+            yield s[0], s[1], sid, auth, addr, s[2], s[3], s[4]
+
+        if sid:
+            klass = self.__class__.__name__
+            for s in Subscription.find_by_sids_and_class(self.env,
+                    ((sid,auth),), klass):
+                yield s.subscription_tuple()
+
+    def description(self):
+        return _("I update a ticket")
+
+    def default_subscriptions(self):
+        klass = self.__class__.__name__
+        return NotificationSystem(self.env).default_subscriptions(klass)
+
+    def requires_authentication(self):
+        return True
+
+
+class TicketPreviousUpdatersSubscriber(Component):
+    """Allows subscribing to future changes simply by updating a ticket."""
+
+    implements(INotificationSubscriber)
+
+    def matches(self, event):
+        if event.realm != 'ticket':
+            return
+        if event.category not in ('created', 'changed', 'attachment added',
+                                  'attachment deleted'):
+            return
+
+        updaters = [row[0] for row in self.env.db_query("""
+            SELECT DISTINCT author FROM ticket_change
+            WHERE ticket=%s
+            """, (event.target.id, ))]
+
+        matcher = RecipientMatcher(self.env)
+        klass = self.__class__.__name__
+        sids = set()
+        for previous_updater in updaters:
+            if previous_updater == event.author:
+                continue
+
+            recipient = matcher.match_recipient(previous_updater)
+            if not recipient:
+                continue
+            sid, auth, addr = recipient
+
+            # Default subscription
+            for s in self.default_subscriptions():
+                yield s[0], s[1], sid, auth, addr, s[2], s[3], s[4]
+            if sid:
+                sids.add((sid,auth))
+
+        for s in Subscription.find_by_sids_and_class(self.env, sids, klass):
+            yield s.subscription_tuple()
+
+    def description(self):
+        return _("Ticket that I previously updated is modified")
+
+    def default_subscriptions(self):
+        klass = self.__class__.__name__
+        return NotificationSystem(self.env).default_subscriptions(klass)
+
+    def requires_authentication(self):
+        return True
+
+
+class TicketReporterSubscriber(Component):
+    """Allows the users to subscribe to tickets that they report."""
+
+    implements(INotificationSubscriber)
+
+    def matches(self, event):
+        if event.realm != 'ticket':
+            return
+        if event.category not in ('created', 'changed', 'attachment added',
+                                  'attachment deleted'):
+            return
+
+        ticket = event.target
+
+        matcher = RecipientMatcher(self.env)
+        recipient = matcher.match_recipient(ticket['reporter'])
+        if not recipient:
+            return
+        sid, auth, addr = recipient
+
+        # Default subscription
+        for s in self.default_subscriptions():
+            yield s[0], s[1], sid, auth, addr, s[2], s[3], s[4]
+
+        if sid:
+            klass = self.__class__.__name__
+            for s in Subscription.find_by_sids_and_class(self.env,
+                    ((sid,auth),), klass):
+                yield s.subscription_tuple()
+
+    def description(self):
+        return _("Ticket that I reported is modified")
+
+    def default_subscriptions(self):
+        klass = self.__class__.__name__
+        return NotificationSystem(self.env).default_subscriptions(klass)
+
+    def requires_authentication(self):
+        return True
+
+
+class NewTicketSubscriber(Component):
+    """Allows the users to subscribe to new tickets."""
+
+    implements(INotificationSubscriber)
+
+    # INotificationSubscriber methods
+
+    def matches(self, event):
+        if event.realm != 'ticket' or event.category != 'created':
+            return
+
+        klass = self.__class__.__name__
+        for s in Subscription.find_by_class(self.env, klass):
+            yield s.subscription_tuple()
+
+    def description(self):
+        return _("Any ticket is created")
+
+    def default_subscriptions(self):
+        return []
+
+    def requires_authentication(self):
+        return False
+
+
+class CarbonCopySubscriber(Component):
+    """Carbon copy subscriber for cc ticket field."""
+
+    implements(INotificationSubscriber)
+
+    def matches(self, event):
+        if event.realm != 'ticket':
+            return
+        if event.category not in ('created', 'changed', 'attachment added',
+                                  'attachment deleted'):
+            return
+
+        # CC field is stored as comma-separated string. Parse to set.
+        chrome = Chrome(self.env)
+        to_set = lambda cc: set(chrome.cc_list(cc))
+        cc_set = to_set(event.target['cc'] or '')
+
+        # Harvest previous CC field
+        if 'fields' in event.changes and 'cc' in event.changes['fields']:
+            cc_set.update(to_set(event.changes['fields']['cc']['old']))
+
+        matcher = RecipientMatcher(self.env)
+        klass = self.__class__.__name__
+        sids = set()
+        for cc in cc_set:
+            recipient = matcher.match_recipient(cc)
+            if not recipient:
+                continue
+            sid, auth, addr = recipient
+
+            # Default subscription
+            for s in self.default_subscriptions():
+                yield s[0], s[1], sid, auth, addr, s[2], s[3], s[4]
+            if sid:
+                sids.add((sid,auth))
+
+        for s in Subscription.find_by_sids_and_class(self.env, sids, klass):
+            yield s.subscription_tuple()
+
+    def description(self):
+        return _("Ticket that I'm listed in the CC field is modified")
+
+    def default_subscriptions(self):
+        klass = self.__class__.__name__
+        return NotificationSystem(self.env).default_subscriptions(klass)
+
+    def requires_authentication(self):
+        return True
+
+
 class TicketNotifyEmail(NotifyEmail):
-    """Notification of ticket changes."""
+    """Notification of ticket changes.
+
+    :since 1.1.3: deprecated and will be removed in 1.3.1
+    """
 
     template_name = "ticket_notify_email.txt"
     from_email = 'trac+ticket@localhost'
@@ -164,6 +540,32 @@ class TicketNotifyEmail(NotifyEmail):
 
     def notify(self, ticket, newticket=True, modtime=None):
         """Send ticket change notification e-mail (untranslated)"""
+        with translation_deactivated(ticket):
+            author = self._prepare_body(ticket, newticket, modtime)
+            subject = self.data['subject']
+            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
+
+    def notify_attachment(self, ticket, attachment, added=True):
+        """Send ticket attachment notification (untranslated)"""
+        with translation_deactivated(ticket):
+            self._prepare_body_attachment(ticket, attachment, added)
+            author = attachment.author
+            subject = self.data['subject']
+            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
+
+    def format(self, ticket, newticket=True, modtime=None):
+        """Format ticket change notification e-mail (untranslated)"""
+        with translation_deactivated(ticket):
+            self._prepare_body(ticket, newticket, modtime)
+            return self._format_body()
+
+    def format_attachment(self, ticket, attachment, added=True):
+        """Format ticket attachment notification e-mail (untranslated)"""
+        with translation_deactivated(ticket):
+            self._prepare_body_attachment(ticket, attachment, added)
+            return self._format_body()
+
+    def _prepare_body(self, ticket, newticket, modtime):
         self.ticket = ticket
         self.modtime = modtime
         self.newticket = newticket
@@ -184,7 +586,7 @@ class TicketNotifyEmail(NotifyEmail):
                     continue
                 author = change['author']
                 change_data.update({
-                    'author': self.obfuscate_email(author),
+                    'author': self.format_author(author),
                     'comment': wrap(change['comment'], self.COLS, ' ', ' ',
                                     '\n', self.ambiwidth)
                 })
@@ -225,8 +627,12 @@ class TicketNotifyEmail(NotifyEmail):
                             changes_body += chgcc
                     else:
                         if field in ['owner', 'reporter']:
-                            old = self.obfuscate_email(old)
-                            new = self.obfuscate_email(new)
+                            old = self.format_author(old)
+                            new = self.format_author(new)
+                        elif field in ticket.time_fields:
+                            format = ticket.fields.by_name(field).get('format')
+                            old = self.format_time_field(old, format)
+                            new = self.format_time_field(new, format)
                         newv = new
                         length = 7 + len(field)
                         spacer_old, spacer_new = ' ', ' '
@@ -260,20 +666,18 @@ class TicketNotifyEmail(NotifyEmail):
 
         subject = self.format_subj(summary, newticket)
 
-        with _translation_deactivated(ticket):
-            self.data.update({
-                'ticket_props': self.format_props(),
-                'ticket_body_hdr': self.format_hdr(),
-                'subject': subject,
-                'ticket': ticket_values,
-                'changes_body': changes_body,
-                'changes_descr': changes_descr,
-                'change': change_data
-            })
-            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
+        self.data.update({
+            'ticket_props': self.format_props(),
+            'ticket_body_hdr': self.format_hdr(),
+            'subject': subject,
+            'ticket': ticket_values,
+            'changes_body': changes_body,
+            'changes_descr': changes_descr,
+            'change': change_data
+        })
+        return author
 
-    def notify_attachment(self, ticket, attachment, added=True):
-        """Send ticket attachment notification (untranslated)"""
+    def _prepare_body_attachment(self, ticket, attachment, added):
         self.ticket = ticket
         self.modtime = attachment.date or datetime_now(utc)
         self.newticket = False
@@ -302,17 +706,16 @@ class TicketNotifyEmail(NotifyEmail):
         ticket_values['new'] = self.newticket
         ticket_values['link'] = link
         subject = self.format_subj(summary, False)
-        with _translation_deactivated(ticket):
-            self.data.update({
-                'ticket_props': self.format_props(),
-                'ticket_body_hdr': self.format_hdr(),
-                'subject': subject,
-                'ticket': ticket_values,
-                'changes_body': changes_body,
-                'changes_descr': '',
-                'change': {'author': self.obfuscate_email(author)},
-            })
-            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
+
+        self.data.update({
+            'ticket_props': self.format_props(),
+            'ticket_body_hdr': self.format_hdr(),
+            'subject': subject,
+            'ticket': ticket_values,
+            'changes_body': changes_body,
+            'changes_descr': '',
+            'change': {'author': self.format_author(author)},
+        })
 
     def format_props(self):
         tkt = self.ticket
@@ -327,10 +730,13 @@ class TicketNotifyEmail(NotifyEmail):
             if fname not in tkt.values:
                 continue
             fval = tkt[fname] or ''
+            if fname in tkt.time_fields:
+                format = tkt.fields.by_name(fname).get('format')
+                fval = self.format_time_field(fval, format)
             if fval.find('\n') != -1:
                 continue
             if fname in ['owner', 'reporter']:
-                fval = self.obfuscate_email(fval)
+                fval = self.format_author(fval)
             idx = 2 * (i % 2)
             width[idx] = max(self.get_text_width(f['label']), width[idx])
             width[idx + 1] = max(self.get_text_width(fval), width[idx + 1])
@@ -360,8 +766,11 @@ class TicketNotifyEmail(NotifyEmail):
             if fname not in tkt.values:
                 continue
             fval = tkt[fname] or ''
+            if fname in tkt.time_fields:
+                format = tkt.fields.by_name(fname).get('format')
+                fval = self.format_time_field(fval, format)
             if fname in ['owner', 'reporter']:
-                fval = self.obfuscate_email(fval)
+                fval = self.format_author(fval)
             if f['type'] == 'textarea' or '\n' in unicode(fval):
                 big.append((f['label'], '\n'.join(fval.splitlines())))
             else:
@@ -418,13 +827,17 @@ class TicketNotifyEmail(NotifyEmail):
         return filter(lambda x: '@' in x, txt.replace(',', ' ').split())
 
     def diff_cc(self, old, new):
-        oldcc = NotifyEmail.addrsep_re.split(old)
-        newcc = NotifyEmail.addrsep_re.split(new)
-        added = [self.obfuscate_email(x)
+        chrome = Chrome(self.env)
+        oldcc = chrome.cc_list(old)
+        newcc = chrome.cc_list(new)
+        added = [self.format_author(x)
                  for x in newcc if x and x not in oldcc]
-        removed = [self.obfuscate_email(x)
-                   for x in oldcc if x and x not in newcc]
-        return added, removed
+        rmved = [self.format_author(x)
+                 for x in oldcc if x and x not in newcc]
+        return added, rmved
+
+    def format_author(self, author):
+        return Chrome(self.env).format_author(None, author)
 
     def format_hdr(self):
         return '#%s: %s' % (self.ticket.id, wrap(self.ticket['summary'],
@@ -451,6 +864,11 @@ class TicketNotifyEmail(NotifyEmail):
             subj = "Re: " + subj
         return subj
 
+    def format_time_field(self, value, format):
+        tzinfo = get_timezone(self.config.get('trac', 'default_timezone'))
+        return format_date_or_datetime(format, value, tzinfo=tzinfo) \
+               if value else ''
+
     def get_recipients(self, tktid):
         to_recipients, cc_recipients, reporter, owner = \
             get_ticket_notification_recipients(self.env, self.config, tktid,
@@ -461,13 +879,9 @@ class TicketNotifyEmail(NotifyEmail):
 
     def get_message_id(self, rcpt, modtime=None):
         """Generate a predictable, but sufficiently unique message ID."""
-        s = '%s.%08d.%d.%s' % (self.env.project_url.encode('utf-8'),
-                               int(self.ticket.id), to_utimestamp(modtime),
-                               rcpt.encode('ascii', 'ignore'))
-        dig = md5(s).hexdigest()
-        host = self.from_email[self.from_email.find('@') + 1:]
-        msgid = '<%03d.%s@%s>' % (len(s), dig, host)
-        return msgid
+        targetid = '%08d' % int(self.ticket.id)
+        return create_message_id(self.env, targetid, self.from_email,
+                                 modtime, rcpt)
 
     def send(self, torcpts, ccrcpts):
         dest = self.reporter or 'anonymous'
@@ -489,6 +903,8 @@ class TicketNotifyEmail(NotifyEmail):
         """ Obfuscate text when `show_email_addresses` is disabled in config.
         Obfuscation happens once per email, regardless of recipients, so
         cannot use permission-based obfuscation.
+
+        :since 1.2: Deprecated and will be removed in 1.3.1.
         """
         if self.env.config.getbool('trac', 'show_email_addresses'):
             return text
@@ -504,10 +920,11 @@ class TicketAttachmentNotifier(Component):
     # IAttachmentChangeListener methods
 
     def attachment_added(self, attachment):
-        self._notify_attachment(attachment, True)
+        self._notify_attachment(attachment, 'attachment added',
+                                attachment.date)
 
     def attachment_deleted(self, attachment):
-        self._notify_attachment(attachment, False)
+        self._notify_attachment(attachment, 'attachment deleted', None)
 
     def attachment_reparented(self, attachment, old_parent_realm,
                               old_parent_id):
@@ -515,15 +932,16 @@ class TicketAttachmentNotifier(Component):
 
     # Internal methods
 
-    def _notify_attachment(self, attachment, added):
+    def _notify_attachment(self, attachment, category, time):
         resource = attachment.resource.parent
         if resource.realm != 'ticket':
             return
         ticket = Ticket(self.env, resource.id)
-        tn = TicketNotifyEmail(self.env)
+        event = TicketChangeEvent(category, ticket, time, ticket['reporter'],
+                                  attachment=attachment)
         try:
-            tn.notify_attachment(ticket, attachment, added)
-        except Exception, e:
+            NotificationSystem(self.env).notify(event)
+        except Exception as e:
             self.log.error("Failure sending notification when adding "
                            "attachment %s to ticket #%s: %s",
                            attachment.filename, ticket.id,
@@ -531,7 +949,10 @@ class TicketAttachmentNotifier(Component):
 
 
 class BatchTicketNotifyEmail(NotifyEmail):
-    """Notification of ticket batch modifications."""
+    """Notification of ticket batch modifications.
+
+    :since 1.1.3: deprecated and will be removed in 1.3.1
+    """
 
     template_name = "batch_ticket_notify_email.txt"
 
@@ -542,10 +963,25 @@ class BatchTicketNotifyEmail(NotifyEmail):
                modtime=None):
         """Send batch ticket change notification e-mail (untranslated)"""
         tickets = self._sort_tickets_by_priority(tickets)
-        with _translation_deactivated():
+        with translation_deactivated():
             self._notify(tickets, new_values, comment, action, author, modtime)
 
     def _notify(self, tickets, new_values, comment, action, author, modtime):
+        self._prepare_body(tickets, new_values, comment, action, author,
+                           modtime)
+        subject = self.data['subject']
+        super(BatchTicketNotifyEmail, self).notify(tickets, subject, author)
+
+    def format(self, tickets, new_values, comment, action, author,
+               modtime=None):
+        """Format batch ticket change notification e-mail (untranslated)"""
+        with translation_deactivated():
+            self._prepare_body(tickets, new_values, comment, action, author,
+                               modtime)
+            return self._format_body()
+
+    def _prepare_body(self, tickets, new_values, comment, action, author,
+                      modtime):
         self.tickets = tickets
         self.reporter = ''
         self.owner = ''
@@ -564,7 +1000,6 @@ class BatchTicketNotifyEmail(NotifyEmail):
             'subject': subject,
             'ticket_query_link': link,
         })
-        super(BatchTicketNotifyEmail, self).notify(tickets, subject, author)
 
     def _sort_tickets_by_priority(self, tickets):
         with self.env.db_query as db:
@@ -608,29 +1043,9 @@ class BatchTicketNotifyEmail(NotifyEmail):
         return list(all_to_recipients), list(all_cc_recipients)
 
     def get_message_id(self, modtime=None):
-        s = '%s.%s.%d' % (self.env.project_url.encode('utf-8'),
-                          ','.join(map(str, self.tickets)),
-                          to_utimestamp(modtime))
-        dig = md5(s).hexdigest()
-        host = self.from_email[self.from_email.find('@') + 1:]
-        msgid = '<%03d.%s@%s>' % (len(s), dig, host)
-        return msgid
+        targetid = ','.join(map(str, self.tickets))
+        return create_message_id(self.env, targetid, self.from_email, modtime)
 
     def send(self, torcpts, ccrcpts):
         hdrs = {'Message-ID': self.get_message_id(self.modtime)}
         super(BatchTicketNotifyEmail, self).send(torcpts, ccrcpts, hdrs)
-
-
-@contextmanager
-def _translation_deactivated(ticket=None):
-    t = deactivate()
-    if ticket is not None:
-        ts = TicketSystem(ticket.env)
-        translated_fields = ticket.fields
-        ticket.fields = ts.get_ticket_fields()
-    try:
-        yield
-    finally:
-        if ticket is not None:
-            ticket.fields = translated_fields
-        reactivate(t)
