@@ -1,3 +1,16 @@
+# -*- coding: utf-8 -*-
+#
+# Copyright (C) 2006-2018 Edgewall Software
+# All rights reserved.
+#
+# This software is licensed as described in the file COPYING, which
+# you should have received as part of this distribution. The terms
+# are also available at http://trac.edgewall.com/license.html.
+#
+# This software consists of voluntary contributions made by many
+# individuals. For the exact contribution history, see the revision
+# history and logs, available at http://trac.edgewall.org/.
+
 from trac.db import Table, Column, Index, DatabaseManager
 
 def do_upgrade(env, ver, cursor):
@@ -28,7 +41,7 @@ def do_upgrade(env, ver, cursor):
                 Column('newvalue'),
                 Index(['ticket']),
                 Index(['time'])]]
-    
+
     db_connector, _ = DatabaseManager(env).get_connector()
     for table in tables:
         for stmt in db_connector.to_sql(table):
@@ -37,25 +50,28 @@ def do_upgrade(env, ver, cursor):
     # Add an index to the temporary table to speed up the conversion
     cursor.execute("CREATE INDEX session_old_sid_idx ON session_old(sid)")
     # Insert the sessions into the new table
-    db = env.get_db_cnx()
-    cursor.execute("INSERT INTO session (sid, last_visit, authenticated) "
-                   "SELECT distinct s.sid,COALESCE(%s,0),s.authenticated "
-                   "FROM session_old AS s LEFT JOIN session_old AS s2 "
-                   "ON (s.sid=s2.sid AND s2.var_name='last_visit') "
-                   "WHERE s.sid IS NOT NULL"
-                   % db.cast('s2.var_value', 'int'))
-    cursor.execute("INSERT INTO session_attribute "
-                   "(sid, authenticated, name, value) "
-                   "SELECT s.sid, s.authenticated, s.var_name, s.var_value "
-                   "FROM session_old s "
-                   "WHERE s.var_name <> 'last_visit' AND s.sid IS NOT NULL")
+    with env.db_query as db:
+        cursor.execute("""
+            INSERT INTO session (sid, last_visit, authenticated)
+            SELECT distinct s.sid,COALESCE(%s,0),s.authenticated
+            FROM session_old AS s LEFT JOIN session_old AS s2
+            ON (s.sid=s2.sid AND s2.var_name='last_visit')
+            WHERE s.sid IS NOT NULL
+            """ % db.cast('s2.var_value', 'int'))
+    cursor.execute("""
+        INSERT INTO session_attribute (sid, authenticated, name, value)
+        SELECT s.sid, s.authenticated, s.var_name, s.var_value
+        FROM session_old s
+        WHERE s.var_name <> 'last_visit' AND s.sid IS NOT NULL
+        """)
 
     # Insert ticket change data into the new table
-    cursor.execute("INSERT INTO ticket_change "
-                   "(ticket, time, author, field, oldvalue, newvalue) "
-                   "SELECT ticket, time, author, field, oldvalue, newvalue "
-                   "FROM ticket_change_old")
+    cursor.execute("""
+        INSERT INTO ticket_change (ticket, time, author, field, oldvalue,
+                                   newvalue)
+        SELECT ticket, time, author, field, oldvalue, newvalue
+        FROM ticket_change_old
+        """)
 
     cursor.execute("DROP TABLE session_old")
     cursor.execute("DROP TABLE ticket_change_old")
-
