@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2006-2013 Edgewall Software
+# Copyright (C) 2006-2018 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -11,7 +11,7 @@
 # individuals. For the exact contribution history, see the revision
 # history and logs, available at http://trac.edgewall.org/.
 
-from trac.db import Table, Column, Index, DatabaseManager
+from trac.db import Table, Column, DatabaseManager
 
 def do_upgrade(env, ver, cursor):
     """Rename the column `sql` in the `report` table for compatibity with MySQL.
