@@ -1,3 +1,16 @@
+# -*- coding: utf-8 -*-
+#
+# Copyright (C) 2010-2018 Edgewall Software
+# All rights reserved.
+#
+# This software is licensed as described in the file COPYING, which
+# you should have received as part of this distribution. The terms
+# are also available at http://trac.edgewall.com/license.html.
+#
+# This software consists of voluntary contributions made by many
+# individuals. For the exact contribution history, see the revision
+# history and logs, available at http://trac.edgewall.org/.
+
 from trac.db import DatabaseManager
 
 
@@ -14,21 +27,21 @@ def do_upgrade(env, ver, cursor):
                        'completed': ('int', 'int64')}),
         ('version', {'time': ('int', 'int64')}),
     ]
-    
+
     db_connector, _ = DatabaseManager(env).get_connector()
-    db = env.get_db_cnx()
     for table, columns in tables:
         # Alter column types
         for sql in db_connector.alter_column_types(table, columns):
             cursor.execute(sql)
-        
+
         # Convert timestamps to microseconds
         cursor.execute("UPDATE %s SET %s" % (table,
                         ', '.join("%s=%s*1000000" % (column, column)
                                   for column in columns)))
-    
+
     # Convert comment edit timestamps to microseconds
-    cursor.execute("UPDATE ticket_change SET newvalue=%s*1000000 "
-                   "WHERE field %s" % (db.cast('newvalue', 'int64'),
-                                       db.like()),
-                   ('_comment%',))
+    with env.db_query as db:
+        cursor.execute("""
+            UPDATE ticket_change SET newvalue=%s*1000000
+            WHERE field %s""" % (db.cast('newvalue', 'int64'), db.like()),
+            ('_comment%',))
