@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2006 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2006 Matthew Good <trac@matt-good.net>
 # Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
@@ -17,28 +17,29 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Matthew Good <trac@matt-good.net>
 
-from __future__ import with_statement
-
 from cStringIO import StringIO
 import csv
 import errno
 import functools
+import hashlib
 import inspect
 from itertools import izip, tee
 import locale
-import os.path
+import os
 from pkg_resources import find_distributions
 import random
 import re
 import shutil
 import sys
+import string
 import struct
 import tempfile
 from urllib import quote, unquote, urlencode
 
-from .compat import any, md5, sha1, sorted
-from .datefmt import time_now, to_datetime, to_timestamp, utc
-from .text import exception_to_unicode, to_unicode, getpreferredencoding
+from trac.util.compat import any, md5, sha1, sorted  # Remove in 1.3.1
+from trac.util.datefmt import time_now, to_datetime, to_timestamp, utc
+from trac.util.text import exception_to_unicode, to_unicode, \
+                           getpreferredencoding
 
 
 def get_reporter_id(req, arg_name=None):
@@ -58,8 +59,8 @@ def get_reporter_id(req, arg_name=None):
         r = req.args.get(arg_name)
         if r:
             return r
-    name = req.session.get('name', None)
-    email = req.session.get('email', None)
+    name = req.session.get('name')
+    email = req.session.get('email')
     if name and email:
         return '%s <%s>' % (name, email)
     return name or email or req.authname # == 'anonymous'
@@ -145,7 +146,7 @@ if os.name == 'nt':
         # Fall back to "move away and replace"
         try:
             os.rename(src, dst)
-        except OSError, e:
+        except OSError as e:
             if e.errno != errno.EEXIST:
                 raise
             old = "%s-%08x" % (dst, random.randint(0, sys.maxint))
@@ -233,10 +234,16 @@ def read_file(path, mode='r'):
 
 
 def create_file(path, data='', mode='w'):
-    """Create a new file with the given data."""
+    """Create a new file with the given data.
+
+    :data: string or iterable of strings.
+    """
     with open(path, mode) as f:
         if data:
-            f.write(data)
+            if isinstance(data, basestring):
+                f.write(data)
+            else:  # Assume iterable
+                f.writelines(data)
 
 
 def create_unique_file(path):
@@ -249,7 +256,7 @@ def create_unique_file(path):
             if hasattr(os, 'O_BINARY'):
                 flags += os.O_BINARY
             return path, os.fdopen(os.open(path, flags, 0666), 'w')
-        except OSError, e:
+        except OSError as e:
             if e.errno != errno.EEXIST:
                 raise
             idx += 1
@@ -274,10 +281,12 @@ else:
         missing."""
         try:
             os.utime(filename, None)
-        except OSError, e:
+        except OSError as e:
             if e.errno == errno.ENOENT:
                 with open(filename, 'ab'):
                     pass
+            else:
+                raise
 
 
 def create_zipinfo(filename, mtime=None, dir=False, executable=False, symlink=False,
@@ -353,7 +362,7 @@ class NaivePopen(object):
 
     Example::
 
-      print Popen3('grep spam','\\n\\nhere spam\\n\\n').out
+      print(Popen3('grep spam','\\n\\nhere spam\\n\\n').out)
     """
     def __init__(self, command, input=None, capturestderr=None):
         outfile = tempfile.mktemp()
@@ -383,35 +392,38 @@ class NaivePopen(object):
 
 
 def terminate(process):
-    """Python 2.5 compatibility method.
-    os.kill is not available on Windows before Python 2.7.
-    In Python 2.6 subprocess.Popen has a terminate method.
-    (It also seems to have some issues on Windows though.)
+    """Terminate the process.
+
+    If the process has already finished and has not been waited for,
+    the function does not raise OSError and WindowsError exceptions unlike
+    a terminate method of `subprocess.Popen`.
+
+    :param process: the integer id (`pid`) of the process.
     """
 
-    def terminate_win(process):
+    pid = process if isinstance(process, int) else process.pid
+
+    def terminate_win():
         import ctypes
         PROCESS_TERMINATE = 1
         handle = ctypes.windll.kernel32.OpenProcess(PROCESS_TERMINATE,
-                                                    False,
-                                                    process.pid)
+                                                    False, pid)
         ctypes.windll.kernel32.TerminateProcess(handle, -1)
         ctypes.windll.kernel32.CloseHandle(handle)
 
-    def terminate_nix(process):
-        import os
+    def terminate_nix():
         import signal
         try:
-            os.kill(process.pid, signal.SIGTERM)
-        except OSError, e:
+            os.kill(pid, signal.SIGTERM)
+        except OSError as e:
             # If the process has already finished and has not been
             # waited for, killing it raises an ESRCH error on Cygwin
             if e.errno != errno.ESRCH:
                 raise
 
     if sys.platform == 'win32':
-        return terminate_win(process)
-    return terminate_nix(process)
+        return terminate_win()
+    return terminate_nix()
 
 
 def makedirs(path, overwrite=False):
@@ -462,17 +474,17 @@ def copytree(src, dst, symlinks=False, skip=[], overwrite=False):
                     remove_if_overwriting(dstname)
                     shutil.copy2(srcname, dstname)
                 # XXX What about devices, sockets etc.?
-            except (IOError, OSError), why:
+            except (IOError, OSError) as why:
                 errors.append((srcname, dstname, str(why)))
             # catch the Error from the recursive copytree so that we can
             # continue with other files
-            except shutil.Error, err:
+            except shutil.Error as err:
                 errors.extend(err.args[0])
         try:
             shutil.copystat(src, dst)
         except WindowsError:
             pass  # Ignore errors due to limited Windows copystat support
-        except OSError, why:
+        except OSError as why:
             errors.append((src, dst, str(why)))
         if errors:
             raise shutil.Error(errors)
@@ -483,8 +495,12 @@ def is_path_below(path, parent):
     """Return True iff `path` is equal to parent or is located below `parent`
     at any level.
     """
-    path = os.path.abspath(path)
-    parent = os.path.abspath(parent)
+    def normalize(path):
+        if os.name == 'nt' and not isinstance(path, unicode):
+            path = path.decode('mbcs')
+        return os.path.normcase(os.path.abspath(path))
+    path = normalize(path)
+    parent = normalize(parent)
     return path == parent or path.startswith(parent + os.sep)
 
 
@@ -532,7 +548,6 @@ def arity(f):
 def get_last_traceback():
     """Retrieve the last traceback as an `unicode` string."""
     import traceback
-    from StringIO import StringIO
     tb = StringIO()
     traceback.print_exc(file=tb)
     return to_unicode(tb.getvalue())
@@ -625,7 +640,7 @@ def safe__import__(module_name):
     already_imported = sys.modules.copy()
     try:
         return __import__(module_name, globals(), locals(), [])
-    except Exception, e:
+    except Exception as e:
         for modname in sys.modules.copy():
             if modname not in already_imported:
                 del(sys.modules[modname])
@@ -643,7 +658,7 @@ def safe_repr(x):
     """
     try:
         return to_unicode(repr(x))
-    except Exception, e:
+    except Exception as e:
         return "<%s object at 0x%X (repr() error: %s)>" % (
             fq_class_name(x), id(x), exception_to_unicode(e))
 
@@ -730,9 +745,16 @@ def get_pkginfo(dist):
     Always returns a dictionary but it will be empty if no Distribution
     instance can be created for the given module.
     """
+    import email
+    import email.errors
     import types
+    from trac.util.translation import _
+
+    def parse_pkginfo(dist, name):
+        return email.message_from_string(to_utf8(dist.get_metadata(name)))
+
     if isinstance(dist, types.ModuleType):
-        def has_resource(dist, resource_name):
+        def has_resource(dist, module, resource_name):
             if dist.location.endswith('.egg'):  # installed by easy_install
                 return dist.has_resource(resource_name)
             if dist.has_metadata('installed-files.txt'):  # installed by pip
@@ -747,6 +769,16 @@ def get_pkginfo(dist):
             if dist.has_metadata('RECORD'):  # *.dist-info/RECORD
                 reader = csv.reader(StringIO(dist.get_metadata('RECORD')))
                 return any(resource_name == row[0] for row in reader)
+            if dist.has_metadata('PKG-INFO'):
+                try:
+                    pkginfo = parse_pkginfo(dist, 'PKG-INFO')
+                    provides = pkginfo.get_all('Provides', ())
+                    names = module.__name__.split('.')
+                    if any('.'.join(names[:n + 1]) in provides
+                           for n in xrange(len(names))):
+                        return True
+                except (IOError, email.Errors.MessageError):
+                    pass
             toplevel = resource_name.split('/')[0]
             if dist.has_metadata('top_level.txt'):
                 return toplevel in dist.get_metadata_lines('top_level.txt')
@@ -761,28 +793,27 @@ def get_pkginfo(dist):
             resource_name += '.py'
         for dist in find_distributions(module_path, only=True):
             if os.path.isfile(module_path) or \
-                    has_resource(dist, resource_name):
+                    has_resource(dist, module, resource_name):
                 break
         else:
             return {}
-    import email
-    from trac.util.translation import _
+
     attrs = ('author', 'author-email', 'license', 'home-page', 'summary',
-             'description', 'version')
+             'name', 'description', 'version')
     info = {}
     def normalize(attr):
         return attr.lower().replace('-', '_')
     metadata = 'METADATA' if dist.has_metadata('METADATA') else 'PKG-INFO'
     try:
-        pkginfo = email.message_from_string(dist.get_metadata(metadata))
+        pkginfo = parse_pkginfo(dist, metadata)
         for attr in [key for key in attrs if key in pkginfo]:
             info[normalize(attr)] = pkginfo[attr]
-    except IOError, e:
+    except IOError as e:
         err = _("Failed to read %(metadata)s file for %(dist)s: %(err)s",
                 metadata=metadata, dist=dist, err=to_unicode(e))
         for attr in attrs:
             info[normalize(attr)] = err
-    except email.Errors.MessageError, e:
+    except email.errors.MessageError as e:
         err = _("Failed to parse %(metadata)s file for %(dist)s: %(err)s",
                 metadata=metadata, dist=dist, err=to_unicode(e))
         for attr in attrs:
@@ -814,7 +845,7 @@ except NotImplementedError:
 
     def urandom(n):
         result = []
-        hasher = sha1(str(os.getpid()) + str(time_now()))
+        hasher = hashlib.sha1(str(os.getpid()) + str(time_now()))
         while len(result) * hasher.digest_size < n:
             hasher.update(str(_entropy.random()))
             result.append(hasher.digest())
@@ -828,6 +859,12 @@ def hex_entropy(digits=32):
     return result[:digits] if len(result) > digits else result
 
 
+def salt(length=2):
+    """Returns a string of `length` random letters and numbers."""
+    return ''.join(random.choice(string.ascii_letters + string.digits + '/.')
+                   for x in range(length))
+
+
 # Original license for md5crypt:
 # Based on FreeBSD src/lib/libcrypt/crypt.c 1.2
 #
@@ -845,10 +882,10 @@ def md5crypt(password, salt, magic='$1$'):
     # /* The password first, since that is what is most unknown */
     # /* Then our magic string */
     # /* Then the raw salt */
-    m = md5(password + magic + salt)
+    m = hashlib.md5(password + magic + salt)
 
     # /* Then just as many characters of the MD5(pw,salt,pw) */
-    mixin = md5(password + salt + password).digest()
+    mixin = hashlib.md5(password + salt + password).digest()
     for i in range(0, len(password)):
         m.update(mixin[i % 16])
 
@@ -866,7 +903,7 @@ def md5crypt(password, salt, magic='$1$'):
 
     # /* and now, just to make sure things don't run too fast */
     for i in range(1000):
-        m2 = md5()
+        m2 = hashlib.md5()
         if i & 1:
             m2.update(password)
         else:
@@ -1170,7 +1207,8 @@ class lazy(object):
         instance.__dict__[self.fn.__name__] = value
 
     def __delete__(self, instance):
-        del instance.__dict__[self.fn.__name__]
+        if self.fn.__name__ in instance.__dict__:
+            del instance.__dict__[self.fn.__name__]
 
 
 # -- algorithmic utilities
@@ -1184,22 +1222,6 @@ def embedded_numbers(s):
     return pieces
 
 
-def pairwise(iterable):
-    """
-    >>> list(pairwise([0, 1, 2, 3]))
-    [(0, 1), (1, 2), (2, 3)]
-
-    .. deprecated :: 0.11
-       if this really needs to be used, rewrite it without izip
-    """
-    a, b = tee(iterable)
-    try:
-        b.next()
-    except StopIteration:
-        pass
-    return izip(a, b)
-
-
 def partition(iterable, order=None):
     """
     >>> partition([(1, "a"), (2, "b"), (3, "a")])
@@ -1232,22 +1254,31 @@ def as_int(s, default, min=None, max=None):
     return value
 
 
-def as_bool(value):
+def as_bool(value, default=False):
     """Convert the given value to a `bool`.
 
-    If `value` is a string, return `True` for any of "yes", "true", "enabled",
-    "on" or non-zero numbers, ignoring case. For non-string arguments, return
-    the argument converted to a `bool`, or `False` if the conversion fails.
+    If `value` is a string, return `True` for any of "yes", "true",
+    "enabled", "on" or non-zero numbers, ignoring case. For non-string
+    arguments, return the argument converted to a `bool`, or `default`
+    if the conversion fails.
+
+    :since 1.2: the `default` argument can be specified.
     """
     if isinstance(value, basestring):
         try:
             return bool(float(value))
         except ValueError:
-            return value.strip().lower() in ('yes', 'true', 'enabled', 'on')
+            value = value.strip().lower()
+            if value in ('yes', 'true', 'enabled', 'on'):
+                return True
+            elif value in ('no', 'false', 'disabled', 'off'):
+                return False
+            else:
+                return default
     try:
         return bool(value)
     except (TypeError, ValueError):
-        return False
+        return default
 
 
 def pathjoin(*args):
@@ -1257,11 +1288,36 @@ def pathjoin(*args):
 
 def to_list(splittable, sep=','):
     """Split a string at `sep` and return a list without any empty items.
+
+    >>> to_list('1,2, 3,4 ')
+    ['1', '2', '3', '4']
+    >>> to_list('1;2; 3;4 ', sep=';')
+    ['1', '2', '3', '4']
+    >>> to_list('')
+    []
+    >>> to_list(None)
+    []
+    >>> to_list([])
+    []
     """
+    if not splittable:
+        return []
     split = [x.strip() for x in splittable.split(sep)]
     return [item for item in split if item]
 
 
+def sub_val(the_list, item_to_remove, item_to_add):
+    """Substitute an item if the item is found in a list, otherwise leave
+    the list unmodified.
+    """
+    try:
+        index = the_list.index(item_to_remove)
+    except ValueError:
+        pass
+    else:
+        the_list[index] = item_to_add
+
+
 # Imports for backward compatibility (at bottom to avoid circular dependencies)
 from trac.core import TracError
 from trac.util.compat import reversed
