@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2006 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2006 Matthew Good <trac@matt-good.net>
 # Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
@@ -434,9 +434,9 @@ def get_date_format_jquery_ui(locale):
 def get_time_format_jquery_ui(locale):
     """Get the time format for the jQuery UI timepicker addon."""
     if locale == 'iso8601':
-        return 'hh:mm:ssz'  # XXX timepicker doesn't support 'ISO_8601'
+        return 'HH:mm:ssZ'
     if babel and locale:
-        values = {'h': 'h', 'hh': 'hh', 'H': 'h', 'HH': 'hh',
+        values = {'h': 'h', 'hh': 'hh', 'H': 'H', 'HH': 'HH',
                   'm': 'm', 'mm': 'mm', 's': 's', 'ss': 'ss'}
         f = get_time_format('medium', locale=locale).format
         if '%(a)s' in f:
@@ -450,17 +450,21 @@ def get_time_format_jquery_ui(locale):
     ampm = format_time(t, '%p', tzinfo=utc)
     if ampm:
         tmpl = tmpl.replace(ampm, 'TT' if ampm[0].isupper() else 'tt', 1)
-    return tmpl.replace('23', 'hh', 1).replace('11', 'hh', 1) \
+    return tmpl.replace('23', 'HH', 1).replace('11', 'hh', 1) \
                .replace('59', 'mm', 1).replace('58', 'ss', 1)
 
 def get_timezone_list_jquery_ui(t=None):
     """Get timezone list for jQuery timepicker addon"""
+    def utcoffset(tz, t):  # in minutes
+        offset = t.astimezone(get_timezone(tz)).utcoffset()
+        return offset.days * 24 * 60 + offset.seconds // 60
+    def label(offset):
+        sign = '-' if offset < 0 else '+'
+        return '%s%02d:%02d' % (sign, abs(offset // 60), offset % 60)
     t = to_datetime(t, utc)
-    zones = set(t.astimezone(get_timezone(tz)).strftime('%z')
-                for tz in all_timezones)
-    return [{'value': 'Z', 'label': '+00:00'} \
-            if zone == '+0000' else zone[:-2] + ':' + zone[-2:]
-            for zone in sorted(zones, key=lambda tz: int(tz))]
+    offsets = set(utcoffset(tz, t) for tz in all_timezones)
+    return [{'value': offset, 'label': label(offset)}
+            for offset in sorted(offsets)]
 
 def get_first_week_day_jquery_ui(req):
     """Get first week day for jQuery date picker"""
@@ -612,6 +616,7 @@ def parse_date(text, tzinfo=None, locale=None, hint='date'):
         formatted_hint = {
             'datetime': get_datetime_format_hint,
             'date': get_date_format_hint,
+            'relative': get_datetime_format_hint,
             'iso8601': lambda l: get_datetime_format_hint('iso8601'),
         }.get(hint, lambda(l): hint)(locale)
         if locale == 'iso8601' and hint in ('date', 'datetime') or \
@@ -793,10 +798,12 @@ def _i18n_parse_date_0(text, order, regexp, period_names, month_names, tzinfo):
     t = tzinfo.localize(datetime(*(values[k] for k in 'yMdhms')))
     return tzinfo.normalize(t)
 
-_REL_TIME_RE = re.compile(
-    r'(\d+\.?\d*)\s*'
-    r'(second|minute|hour|day|week|month|year|[hdwmy])s?\s*'
-    r'(?:ago)?$')
+_REL_FUTURE_RE = re.compile(
+    r'(?:in|\+)\s*(\d+\.?\d*)\s*'
+    r'(second|minute|hour|day|week|month|year|[hdwmy])s?$')
+_REL_PAST_RE = re.compile(
+    r'(?:-\s*)?(\d+\.?\d*)\s*'
+    r'(second|minute|hour|day|week|month|year|[hdwmy])s?\s*(?:ago)?$')
 _time_intervals = dict(
     second=lambda v: timedelta(seconds=v),
     minute=lambda v: timedelta(minutes=v),
@@ -811,7 +818,7 @@ _time_intervals = dict(
     m=lambda v: timedelta(days=30 * v),
     y=lambda v: timedelta(days=365 * v),
 )
-_TIME_START_RE = re.compile(r'(this|last)\s*'
+_TIME_START_RE = re.compile(r'(this|last|next)\s*'
                             r'(second|minute|hour|day|week|month|year)$')
 _time_starts = dict(
     second=lambda now: datetime(now.year, now.month, now.day, now.hour,
@@ -837,8 +844,15 @@ def _parse_relative_time(text, tzinfo, now=None):
         dt = _time_starts['day'](now)
     elif text == 'yesterday':
         dt = _time_starts['day'](now) - timedelta(days=1)
+    elif text == 'tomorrow':
+        dt = _time_starts['day'](now) + timedelta(days=1)
     if dt is None:
-        match = _REL_TIME_RE.match(text)
+        match = _REL_FUTURE_RE.match(text)
+        if match:
+            (value, interval) = match.groups()
+            dt = now + _time_intervals[interval](float(value))
+    if dt is None:
+        match = _REL_PAST_RE.match(text)
         if match:
             (value, interval) = match.groups()
             dt = now - _time_intervals[interval](float(value))
@@ -857,6 +871,16 @@ def _parse_relative_time(text, tzinfo, now=None):
                     dt = dt.replace(year=dt.year - 1)
                 else:
                     dt -= _time_intervals[start](1)
+            elif which == 'next':
+                if start == 'month':
+                    if dt.month < 12:
+                        dt = dt.replace(month=dt.month + 1)
+                    else:
+                        dt = dt.replace(year=dt.year + 1, month=1)
+                elif start == 'year':
+                    dt = dt.replace(year=dt.year + 1)
+                else:
+                    dt += _time_intervals[start](1)
 
     if dt is None:
         return None
@@ -864,6 +888,7 @@ def _parse_relative_time(text, tzinfo, now=None):
         dt = tzinfo.localize(dt)
     return tzinfo.normalize(dt)
 
+
 # -- formatting/parsing helper functions
 
 def user_time(req, func, *args, **kwargs):
@@ -883,6 +908,12 @@ def user_time(req, func, *args, **kwargs):
         kwargs['locale'] = getattr(req, 'lc_time', None)
     return func(*args, **kwargs)
 
+def format_date_or_datetime(format, *args, **kwargs):
+    if format == 'date':
+        return format_date(*args, **kwargs)
+    else:
+        return format_datetime(*args, **kwargs)
+
 # -- timezone utilities
 
 class FixedOffset(tzinfo):
@@ -991,11 +1022,11 @@ class LocalTimezone(tzinfo):
             if not std_correct:
                 raise ValueError('Non existent time "%s"' % dt)
         tt = None
-        if std_correct is dst_correct is True:
+        if std_correct and dst_correct:
             tt = local_tt[bool(is_dst)]
-        elif std_correct is True:
+        elif std_correct:
             tt = local_tt[0]
-        elif dst_correct is True:
+        elif dst_correct:
             tt = local_tt[1]
         if tt:
             utc_ts = to_timestamp(datetime(tzinfo=utc, *tt[:6]))
@@ -1003,10 +1034,13 @@ class LocalTimezone(tzinfo):
         else:
             dt = dt.replace(tzinfo=utc)
             utc_ts = to_timestamp(dt)
-            dt -= timedelta(hours=6)
+            dt -= timedelta(seconds=21600)
             tt = (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
                   dt.weekday(), 0, -1)
-            tz_offset = timedelta(seconds=utc_ts - time.mktime(tt) - 6 * 3600)
+            try:
+                tz_offset = timedelta(seconds=utc_ts - time.mktime(tt) - 21600)
+            except (ValueError, OverflowError):
+                return self._std_tz
 
         # if UTC offset doesn't match timezone offset, create a
         # LocalTimezone instance with the UTC offset (#11563)
@@ -1025,7 +1059,13 @@ class LocalTimezone(tzinfo):
         return False
 
     def utcoffset(self, dt):
-        return self._tzinfo(dt)._offset
+        offset = self._tzinfo(dt)._offset
+        if offset.seconds % 60 != 0:
+            # Avoid "ValueError: tzinfo.utcoffset() must return a whole
+            # number of minutes" (#12617)
+            seconds = offset.days * 86400 + offset.seconds
+            offset = timedelta(seconds=int((seconds + 30) // 60) * 60)
+        return offset
 
     def dst(self, dt):
         if self._is_dst(dt):
@@ -1102,7 +1142,8 @@ _timezones = [
     FixedOffset(360, 'GMT +6:00'),   FixedOffset(420, 'GMT +7:00'),
     FixedOffset(480, 'GMT +8:00'),   FixedOffset(540, 'GMT +9:00'),
     FixedOffset(600, 'GMT +10:00'),  FixedOffset(660, 'GMT +11:00'),
-    FixedOffset(720, 'GMT +12:00'),  FixedOffset(780, 'GMT +13:00')]
+    FixedOffset(720, 'GMT +12:00'),  FixedOffset(780, 'GMT +13:00'),
+    FixedOffset(840, 'GMT +14:00')]
 _tzmap = dict([(z.zone, z) for z in _timezones])
 
 all_timezones = [z.zone for z in _timezones]
