@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2018 Edgewall Software
 # Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2006 Matthew Good <trac@matt-good.net>
 # Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
@@ -27,9 +27,6 @@ import textwrap
 from urllib import quote, quote_plus, unquote
 from unicodedata import east_asian_width
 
-from trac.util.translation import _
-
-
 CRLF = '\r\n'
 
 class Empty(unicode):
@@ -228,19 +225,15 @@ def quote_query_string(text):
 
 
 def to_utf8(text, charset='latin1'):
-    """Convert a string to an UTF-8 `str` object.
+    """Convert input to a UTF-8 `str` object.
 
     If the input is not an `unicode` object, we assume the encoding is
     already UTF-8, ISO Latin-1, or as specified by the optional
     *charset* parameter.
     """
-    if isinstance(text, unicode):
-        u = text
-    else:
+    if isinstance(text, str):
         try:
-            # Do nothing if it's already utf-8
             u = unicode(text, 'utf-8')
-            return text
         except UnicodeError:
             try:
                 # Use the user supplied charset if possible
@@ -248,6 +241,11 @@ def to_utf8(text, charset='latin1'):
             except UnicodeError:
                 # This should always work
                 u = unicode(text, 'latin1')
+        else:
+            # Do nothing if it's already utf-8
+            return text
+    else:
+        u = to_unicode(text)
     return u.encode('utf-8')
 
 
@@ -327,26 +325,32 @@ def text_width(text, ambiwidth=1):
     return sum([2 if east_asian_width(chr) in twice else 1
                 for chr in to_unicode(text)])
 
-_default_ambiwidth = 1  # Default width of East Asian Ambiguous (A)
-if os.name == 'nt':
-    try:
-        # `ctypes` is available since Python 2.5
+
+def _get_default_ambiwidth():
+    """Return width of East Asian Ambiguous based on locale environment
+    variables or Windows codepage.
+    """
+
+    if os.name == 'nt':
         import ctypes
         codepage = ctypes.windll.kernel32.GetConsoleOutputCP()
-    except ImportError:
-        # Try to retrieve the codepage from stderr and stdout
-        codepage = (sys.stderr.encoding or sys.stdout.encoding or '')[2:]
-        codepage = codepage.isdigit() and int(codepage) or 0
+        if codepage in (932,   # Japanese (Shift-JIS)
+                        936,   # Chinese Simplified (GB2312)
+                        949,   # Korean (Unified Hangul Code)
+                        950):  # Chinese Traditional (Big5)
+            return 2
+    else:
+        for name in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
+            value = os.environ.get(name) or ''
+            if value:
+                if name == 'LANGUAGE' and ':' in value:
+                    value = value.split(':')[0]
+                return 2 if value.lower().startswith(('zh', 'ja', 'ko')) else 1
 
-    if codepage in (932,  # Japanese (Shift-JIS)
-                    936,  # Chinese Simplified (GB2312)
-                    949,  # Korean (Unified Hangul Code)
-                    950): # Chinese Traditional (Big5)
-        _default_ambiwidth = 2
-    del codepage
-else:
-    if re.match(r'zh|ja|kr', os.environ.get('LANG') or '', re.IGNORECASE):
-        _default_ambiwidth = 2
+    return 1
+
+
+_default_ambiwidth = _get_default_ambiwidth()
 
 
 def print_table(data, headers=None, sep='  ', out=None, ambiwidth=None):
@@ -382,8 +386,23 @@ def print_table(data, headers=None, sep='  ', out=None, ambiwidth=None):
     def tw(text):
         return text_width(text, ambiwidth=ambiwidth)
 
-    # Convert each cell to an unicode object
-    data = [[to_text(cell) for cell in row] for row in data]
+    def to_lines(data):
+        lines = []
+        for row in data:
+            row = [to_text(cell) for cell in row]
+            if any('\n' in cell for cell in row):
+                row = [cell.splitlines() for cell in row]
+                max_lines = max(len(cell) for cell in row)
+                for cell in row:
+                    if len(cell) < max_lines:
+                        cell += [''] * (max_lines - len(cell))
+                lines.extend([cell[idx] for cell in row]
+                             for idx in xrange(max_lines))
+            else:
+                lines.append(row)
+        return lines
+
+    data = to_lines(data)
 
     num_cols = len(data[0])
     col_width = [max(tw(row[idx]) for row in data)
@@ -392,15 +411,15 @@ def print_table(data, headers=None, sep='  ', out=None, ambiwidth=None):
     out.write('\n')
     for ridx, row in enumerate(data):
         for cidx, cell in enumerate(row):
-            if headers and ridx == 0:
-                sp = '%*s' % (tw(sep), ' ') # No separator in header
-            else:
-                sp = sep
             if cidx + 1 == num_cols:
-                sp = '' # No separator after last column
-
-            line = u'%-*s%s' % (col_width[cidx] - tw(cell) + len(cell),
-                                cell, sp)
+                line = cell  # No separator after last column
+            else:
+                if headers and ridx == 0:
+                    sp = ' ' * tw(sep)  # No separator in header
+                else:
+                    sp = sep
+                line = u'%-*s%s' % (col_width[cidx] - tw(cell) + len(cell),
+                                    cell, sp)
             line = line.encode(charset, 'replace')
             out.write(line)
 
@@ -573,6 +592,8 @@ def wrap(t, cols=75, initial_indent='', subsequent_indent='',
     return linesep.join(wrappedLines)
 
 
+_obfuscation_char = u'@\u2026'
+
 def obfuscate_email_address(address):
     """Replace anything looking like an e-mail address (``'@something'``)
     with a trailing ellipsis (``'@…'``)
@@ -580,11 +601,20 @@ def obfuscate_email_address(address):
     if address:
         at = address.find('@')
         if at != -1:
-            return address[:at] + u'@\u2026' + \
+            return address[:at] + _obfuscation_char + \
                    ('>' if address[-1] == '>' else '')
     return address
 
 
+def is_obfuscated(word):
+    """Returns `True` if the `word` looks like an obfuscated e-mail
+    address.
+
+    :since: 1.2
+    """
+    return _obfuscation_char in word
+
+
 def breakable_path(path):
     """Make a path breakable after path separators, and conversely, avoid
     breaking at spaces.
@@ -610,6 +640,7 @@ def normalize_whitespace(text, to_space=u'\u00a0', remove=u'\u200b'):
         text = text.replace(each, '')
     return text
 
+
 def unquote_label(txt):
     """Remove (one level of) enclosing single or double quotes.
 
@@ -617,6 +648,13 @@ def unquote_label(txt):
     """
     return txt[1:-1] if txt and txt[0] in "'\"" and txt[0] == txt[-1] else txt
 
+
+def cleandoc(message):
+    """Removes uniform indentation and leading/trailing whitespace."""
+    from inspect import cleandoc
+    return cleandoc(message).strip()
+
+
 # -- Conversion
 
 def pretty_size(size, format='%.1f'):
@@ -630,6 +668,7 @@ def pretty_size(size, format='%.1f'):
 
     jump = 1024
     if size < jump:
+        from trac.util.translation import _
         return _('%(size)s bytes', size=size)
 
     units = ['KB', 'MB', 'GB', 'TB']
@@ -717,3 +756,17 @@ def levenshtein_distance(lhs, rhs):
                             prev[ridx] + cost)) # substitution
         prev = curr
     return prev[-1]
+
+
+sub_vars_re = re.compile("[$]([A-Z_][A-Z0-9_]*)")
+
+def sub_vars(text, args):
+    """Substitute $XYZ-style variables in a string with provided values.
+
+    :param text: string containing variables to substitute.
+    :param args: dictionary with keys matching the variables to be substituted.
+                 The keys should not be prefixed with the $ character."""
+    def repl(match):
+        key = match.group(1)
+        return args[key] if key in args else '$' + key
+    return sub_vars_re.sub(repl, text)
