@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2009 Edgewall Software
+# Copyright (C) 2005-2018 Edgewall Software
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -14,8 +14,6 @@
 #
 # Author: Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import with_statement
-
 import os
 
 from trac.cache import cached
@@ -42,6 +40,10 @@ CACHE_YOUNGEST_REV = 'youngest_rev'
 CACHE_METADATA_KEYS = (CACHE_REPOSITORY_DIR, CACHE_YOUNGEST_REV)
 
 
+def _norm_reponame(repos):
+    return repos.reponame or '(default)'
+
+
 class CachedRepository(Repository):
 
     has_linear_changesets = False
@@ -89,19 +91,17 @@ class CachedRepository(Repository):
         old_cset = None
 
         with self.env.db_transaction as db:
-            for time, author, message in db("""
-                    SELECT time, author, message FROM revision
-                    WHERE repos=%s AND rev=%s
-                    """, (self.id, srev)):
-                old_cset = Changeset(self.repos, cset.rev, message, author,
-                                     from_utimestamp(time))
+            try:
+                old_cset = CachedChangeset(self, cset.rev, self.env)
+            except NoSuchChangeset:
+                old_cset = None
             if old_cset:
                 db("""UPDATE revision SET time=%s, author=%s, message=%s
                       WHERE repos=%s AND rev=%s
                       """, (to_utimestamp(cset.date), cset.author,
                             cset.message, self.id, srev))
             else:
-                self._insert_changeset(db, cset.rev, cset)
+                self.insert_changeset(cset.rev, cset)
         return old_cset
 
     @cached('_metadata_id')
@@ -130,18 +130,20 @@ class CachedRepository(Repository):
         if youngest:
             youngest = self.repos.normalize_rev(youngest)
             if not youngest:
-                self.log.debug('normalize_rev failed (youngest_rev=%r)',
-                               self.youngest_rev)
+                self.log.debug("normalize_rev failed (youngest_rev=%r, "
+                               "reponame=%s)",
+                               self.youngest_rev, _norm_reponame(self))
         else:
-            self.log.debug('cache metadata undefined (youngest_rev=%r)',
-                           self.youngest_rev)
+            self.log.debug("cache metadata undefined (youngest_rev=%r, "
+                           "reponame=%s)",
+                           self.youngest_rev, _norm_reponame(self))
             youngest = None
 
         # -- compare them and try to resync if different
         next_youngest = None
         if youngest != repos_youngest:
-            self.log.info("repos rev [%s] != cached rev [%s]",
-                          repos_youngest, youngest)
+            self.log.info("repos rev [%s] != cached rev [%s] in '%s'",
+                          repos_youngest, youngest, _norm_reponame(self))
             if youngest:
                 next_youngest = self.repos.next_rev(youngest)
             else:
@@ -178,15 +180,22 @@ class CachedRepository(Repository):
                 srev = self.db_rev(next_youngest)
 
                 with self.env.db_transaction as db:
-                    self.log.info("Trying to sync revision [%s]",
-                                  next_youngest)
+                    self.log.info("Trying to sync revision [%s] in '%s'",
+                                  next_youngest, _norm_reponame(self))
                     cset = self.repos.get_changeset(next_youngest)
                     try:
                         # steps 1. and 2.
-                        self._insert_changeset(db, next_youngest, cset)
-                    except Exception, e: # *another* 1.1. resync attempt won
-                        self.log.warning('Revision %s already cached: %r',
-                                         next_youngest, e)
+                        self.insert_changeset(next_youngest, cset)
+                    except Exception as e: # *another* 1.1. resync attempt won
+                        if isinstance(e, self.env.db_exc.IntegrityError):
+                            self.log.warning("Revision %s in '%s' already "
+                                             "cached: %r", next_youngest,
+                                             _norm_reponame(self), e)
+                        else:
+                            self.log.error("Unable to create cache records "
+                                           "for revision %s in '%s': %r",
+                                           next_youngest, _norm_reponame(self),
+                                           e)
                         # the other resync attempts is also
                         # potentially still in progress, so for our
                         # process/thread, keep ''previous'' notion of
@@ -213,7 +222,7 @@ class CachedRepository(Repository):
 
     def remove_cache(self):
         """Remove the repository cache."""
-        self.log.info("Cleaning cache")
+        self.log.info("Cleaning cache in '%s'", _norm_reponame(self))
         with self.env.db_transaction as db:
             db("DELETE FROM revision WHERE repos=%s",
                (self.id,))
@@ -246,7 +255,7 @@ class CachedRepository(Repository):
                                       "you should resynchronize the "
                                       "repository with: trac-admin $ENV "
                                       "repository resync '%(reponame)s'",
-                                      reponame=self.reponame or '(default)'))
+                                      reponame=_norm_reponame(self)))
             elif repository_dir is None: #
                 self.log.info('Storing initial "repository_dir": %s',
                               self.name)
@@ -270,7 +279,16 @@ class CachedRepository(Repository):
             if invalidate:
                 del self.metadata
 
+    def insert_changeset(self, rev, cset):
+        """Create revision and node_change records for the given changeset
+        instance."""
+        with self.env.db_transaction as db:
+            self._insert_changeset(db, rev, cset)
+
     def _insert_changeset(self, db, rev, cset):
+        """:deprecated: since 1.1.2, use `insert_changeset` instead. Will
+                        be removed in 1.3.1.
+        """
         srev = self.db_rev(rev)
         # 1. Attempt to resync the 'revision' table.  In case of
         # concurrent syncs, only such insert into the `revision` table
@@ -283,7 +301,8 @@ class CachedRepository(Repository):
         # 2. now *only* one process was able to get there (i.e. there
         # *shouldn't* be any race condition here)
         for path, kind, action, bpath, brev in cset.get_changes():
-            self.log.debug("Caching node change in [%s]: %r", rev,
+            self.log.debug("Caching node change in [%s] in '%s': %r",
+                           rev, _norm_reponame(self.repos),
                            (path, kind, action, bpath, brev))
             kind = _inverted_kindmap[kind]
             action = _inverted_actionmap[action]
@@ -332,46 +351,46 @@ class CachedRepository(Repository):
                                                      in node_infos)
         path_revs = dict((node.path, []) for node, first in node_infos)
 
-        db = self.env.get_read_db()
-        cursor = db.cursor()
-        prefix_match = db.prefix_match()
-
         # Prevent "too many SQL variables" since max number of parameters is
         # 999 on SQLite. No limitation on PostgreSQL and MySQL.
         idx = 0
         delta = (999 - 3) // 5
-        while idx < len(node_infos):
-            subset = node_infos[idx:idx + delta]
-            idx += delta
-            count = len(subset)
+        with self.env.db_query as db:
+            prefix_match = db.prefix_match()
+            while idx < len(node_infos):
+                subset = node_infos[idx:idx + delta]
+                idx += delta
+                count = len(subset)
 
-            holders = ','.join(('%s',) * count)
-            query = """\
-                SELECT DISTINCT
-                  rev, (CASE WHEN path IN (%s) THEN path %s END) AS path
-                FROM node_change
-                WHERE repos=%%s AND rev>=%%s AND rev<=%%s AND (path IN (%s) %s)
-                """ % \
-                (holders,
-                 ' '.join(('WHEN path ' + prefix_match + ' THEN %s',) * count),
-                 holders,
-                 ' '.join(('OR path ' + prefix_match,) * count))
-            args = []
-            args.extend(node.path for node, first in subset)
-            for node, first in subset:
-                args.append(db.prefix_match_value(node.path + '/'))
-                args.append(node.path)
-            args.extend((self.id, sfirst, slast))
-            args.extend(node.path for node, first in subset)
-            args.extend(db.prefix_match_value(node.path + '/')
-                        for node, first in subset)
-            cursor.execute(query, args)
+                holders = ','.join(('%s',) * count)
+                query = """\
+                    SELECT DISTINCT
+                      rev, (CASE WHEN path IN (%s) THEN path %s END) AS path
+                    FROM node_change
+                    WHERE repos=%%s AND rev>=%%s AND rev<=%%s
+                      AND (path IN (%s) %s)
+                    """ % \
+                    (holders,
+                     ' '.join(('WHEN path ' + prefix_match + ' THEN %s',)
+                              * count),
+                     holders,
+                     ' '.join(('OR path ' + prefix_match,)
+                              * count))
+                args = []
+                args.extend(node.path for node, first in subset)
+                for node, first in subset:
+                    args.append(db.prefix_match_value(node.path + '/'))
+                    args.append(node.path)
+                args.extend((self.id, sfirst, slast))
+                args.extend(node.path for node, first in subset)
+                args.extend(db.prefix_match_value(node.path + '/')
+                            for node, first in subset)
 
-            for srev, path in cursor:
-                rev = self.rev_db(srev)
-                node, first = path_infos[path]
-                if first <= rev <= node.rev:
-                    path_revs[path].append(rev)
+                for srev, path in db(query, args):
+                    rev = self.rev_db(srev)
+                    node, first = path_infos[path]
+                    if first <= rev <= node.rev:
+                        path_revs[path].append(rev)
 
         return path_revs
 
@@ -479,15 +498,18 @@ class CachedChangeset(Changeset):
 
     def __init__(self, repos, rev, env):
         self.env = env
+        drev = repos.db_rev(rev)
         for _date, author, message in self.env.db_query("""
                 SELECT time, author, message FROM revision
                 WHERE repos=%s AND rev=%s
-                """, (repos.id, repos.db_rev(rev))):
+                """, (repos.id, drev)):
             date = from_utimestamp(_date)
             Changeset.__init__(self, repos, repos.rev_db(rev), message, author,
                                date)
             break
         else:
+            repos.log.debug("Missing revision record (%r, %r) in '%s'",
+                            repos.id, drev, _norm_reponame(repos))
             raise NoSuchChangeset(rev)
 
     def get_changes(self):
