@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2009 Edgewall Software
+# Copyright (C) 2004-2018 Edgewall Software
 # Copyright (C) 2004-2007 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005 Matthew Good <trac@matt-good.net>
 # All rights reserved.
@@ -16,8 +16,6 @@
 # Author: Christopher Lenz <cmlenz@gmx.de>
 #         Matthew Good <trac@matt-good.net>
 
-from __future__ import with_statement
-
 import os
 import pkg_resources
 import sys
@@ -29,7 +27,10 @@ except ImportError:
 
 from mod_python import apache
 try:
-    from mod_python import version
+    try:
+        from mod_python import mp_version as version
+    except ImportError:
+        from mod_python import version
 except ImportError:
     version = "< 3.2"
 
@@ -111,7 +112,7 @@ class ModPythonGateway(WSGIGateway):
         self._send_headers()
         try:
             self.req.sendfile(fileobj.name)
-        except IOError, e:
+        except IOError as e:
             if 'client closed connection' not in str(e):
                 raise
 
@@ -119,7 +120,7 @@ class ModPythonGateway(WSGIGateway):
         self._send_headers()
         try:
             self.req.write(data)
-        except IOError, e:
+        except IOError as e:
             if 'client closed connection' not in str(e):
                 raise
 
