@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2009 Edgewall Software
+# Copyright (C) 2004-2018 Edgewall Software
 # Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2004-2006 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2006 Jonas Borgström <jonas@edgewall.com>
@@ -18,22 +18,21 @@
 # Author: Daniel Lundin <daniel@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import with_statement
-
 import re
-import sys
 
-from trac.admin.api import console_date_format, get_console_locale
-from trac.core import TracError, Component, implements
-from trac.util import hex_entropy
-from trac.util.text import print_table
-from trac.util.translation import _
+from trac.admin.api import AdminCommandError, IAdminCommandProvider, \
+                           console_date_format, get_console_locale
+from trac.core import Component, ExtensionPoint, TracError, TracValueError, \
+                      implements
+from trac.util import as_bool, as_int, hex_entropy, lazy
 from trac.util.datefmt import get_datetime_format_hint, format_date, \
                               parse_date, time_now, to_datetime, to_timestamp
-from trac.admin.api import IAdminCommandProvider, AdminCommandError
+from trac.util.text import print_table
+from trac.util.translation import _
+from trac.web.api import IRequestHandler, is_valid_default_handler
 
 UPDATE_INTERVAL = 3600 * 24 # Update session last_visit time stamp after 1 day
-PURGE_AGE = 3600 * 24 * 90 # Purge session after 90 days idle
+PURGE_AGE = 3600 * 24 * 90 # Expire cookie after 90 days
 COOKIE_KEY = 'trac_session'
 
 
@@ -43,21 +42,50 @@ COOKIE_KEY = 'trac_session'
 #       accessing the tables in the same order within the transaction,
 #       first `session`, then `session_attribute`.
 
-class DetachedSession(dict):
-    def __init__(self, env, sid):
-        dict.__init__(self)
-        self.env = env
+class SessionDict(dict):
+
+    def __init__(self, *args, **kwargs):
+        super(SessionDict, self).__init__(*args, **kwargs)
+        self.authenticated = False
+        self.last_visit = 0
         self.sid = None
-        if sid:
-            self.get_session(sid, authenticated=True)
-        else:
-            self.authenticated = False
-            self.last_visit = 0
-            self._new = True
-            self._old = {}
 
     def __setitem__(self, key, value):
-        dict.__setitem__(self, key, unicode(value))
+        super(SessionDict, self).__setitem__(key, unicode(value))
+
+    def as_bool(self, key, default=None):
+        """Return the value as a boolean. Return `default` if
+        if an exception is raised while converting the value to a
+        boolean.
+
+        :param key: the name of the session attribute
+        :keyword default: the value to return if the parameter is not
+                          specified or an exception occurs converting
+                          the value to a boolean.
+
+        :since: 1.2
+        """
+        if key not in self:
+            return default
+        return as_bool(self[key], default)
+
+    def as_int(self, key, default=None, min=None, max=None):
+        """Return the value as an integer. Return `default` if
+        if an exception is raised while converting the value to an
+        integer.
+
+        :param key: the name of the session attribute
+        :keyword default: the value to return if the parameter is not
+                          specified or an exception occurs converting
+                          the value to an integer.
+        :keyword min: lower bound to which the value is limited
+        :keyword max: upper bound to which the value is limited
+
+        :since: 1.2
+        """
+        if key not in self:
+            return default
+        return as_int(self[key], default, min, max)
 
     def set(self, key, value, default=None):
         """Set a variable in the session, or remove it if it's equal to the
@@ -69,7 +97,18 @@ class DetachedSession(dict):
             if value == default:
                 self.pop(key, None)
                 return
-        dict.__setitem__(self, key, value)
+        super(SessionDict, self).__setitem__(key, value)
+
+
+class DetachedSession(SessionDict):
+
+    def __init__(self, env, sid):
+        super(DetachedSession, self).__init__()
+        self.env = env
+        self._new = True
+        self._old = {}
+        if sid:
+            self.get_session(sid, authenticated=True)
 
     def get_session(self, sid, authenticated=False):
         self.env.log.debug("Retrieving session for ID %r", sid)
@@ -97,14 +136,13 @@ class DetachedSession(dict):
                 self._old = {}
 
     def save(self):
-        items = self.items()
-        if not self._old and not items:
-            # The session doesn't have associated data, so there's no need to
-            # persist it
-            return
-
         authenticated = int(self.authenticated)
         now = int(time_now())
+        items = self.items()
+        if not authenticated and not self._old and not items:
+            # The session for anonymous doesn't have associated data,
+            # so there's no need to persist it
+            return
 
         # We can't do the session management in one big transaction,
         # as the intertwined changes to both the session and
@@ -119,7 +157,8 @@ class DetachedSession(dict):
             # do so is not critical but we nevertheless skip the
             # following steps.
 
-            if self._new:
+            new = self._new
+            if new:
                 self.last_visit = now
                 self._new = False
                 # The session might already exist even if _new is True since
@@ -133,13 +172,15 @@ class DetachedSession(dict):
                     db.rollback()
                     return
 
+            if authenticated and \
+                    (new or self._old.get('name') != self.get('name') or \
+                     self._old.get('email') != self.get('email')):
+                self.env.invalidate_known_users_cache()
+
             # Remove former values for session_attribute and save the
             # new ones. The last concurrent request to do so "wins".
 
             if self._old != self:
-                if self._old.get('name') != self.get('name') or \
-                        self._old.get('email') != self.get('email'):
-                    self.env.invalidate_known_users_cache()
                 if not items and not authenticated:
                     # No need to keep around empty unauthenticated sessions
                     db("DELETE FROM session WHERE sid=%s AND authenticated=0",
@@ -169,7 +210,8 @@ class DetachedSession(dict):
 
         if session_saved and now - self.last_visit > UPDATE_INTERVAL:
             self.last_visit = now
-            mintime = now - PURGE_AGE
+            lifetime = self.env.anonymous_session_lifetime
+            mintime = now - lifetime * 86400 if lifetime > 0 else None
 
             with self.env.db_transaction as db:
                 # Update the session last visit time if it is over an
@@ -178,21 +220,23 @@ class DetachedSession(dict):
                 db("""UPDATE session SET last_visit=%s
                       WHERE sid=%s AND authenticated=%s
                       """, (self.last_visit, self.sid, authenticated))
-                self.env.log.debug('Purging old, expired, sessions.')
-                db("""DELETE FROM session_attribute
-                      WHERE authenticated=0 AND sid IN (
-                          SELECT sid FROM session
-                          WHERE authenticated=0 AND last_visit < %s
-                      )
-                      """, (mintime,))
+                if mintime:
+                    self.env.log.debug('Purging old, expired, sessions.')
+                    db("""DELETE FROM session_attribute
+                          WHERE authenticated=0 AND sid IN (
+                              SELECT sid FROM session
+                              WHERE authenticated=0 AND last_visit < %s
+                          )
+                          """, (mintime,))
 
             # Avoid holding locks on lot of rows on both session_attribute
             # and session tables
-            with self.env.db_transaction as db:
-                db("""
-                    DELETE FROM session
-                    WHERE authenticated=0 AND last_visit < %s
-                    """, (mintime,))
+            if mintime:
+                with self.env.db_transaction as db:
+                    db("""
+                        DELETE FROM session
+                        WHERE authenticated=0 AND last_visit < %s
+                        """, (mintime,))
 
 
 class Session(DetachedSession):
@@ -221,8 +265,7 @@ class Session(DetachedSession):
         self.req.outcookie[COOKIE_KEY]['expires'] = expires
         if self.env.secure_cookies:
             self.req.outcookie[COOKIE_KEY]['secure'] = True
-        if sys.version_info >= (2, 6):
-            self.req.outcookie[COOKIE_KEY]['httponly'] = True
+        self.req.outcookie[COOKIE_KEY]['httponly'] = True
 
     _valid_sid_re = re.compile(r'[_A-Za-z0-9]+\Z')
 
@@ -230,7 +273,7 @@ class Session(DetachedSession):
         refresh_cookie = False
 
         if not authenticated and not self._valid_sid_re.match(sid):
-            raise TracError(_("Session ID must be alphanumeric."))
+            raise TracValueError(_("Session ID must be alphanumeric."))
         if self.sid and sid != self.sid:
             refresh_cookie = True
 
@@ -249,8 +292,8 @@ class Session(DetachedSession):
         if new_sid == self.sid:
             return
         if not self._valid_sid_re.match(new_sid):
-            raise TracError(_("Session ID must be alphanumeric."),
-                            _("Error renaming session"))
+            raise TracValueError(_("Session ID must be alphanumeric."),
+                                 _("Error renaming session"))
         with self.env.db_transaction as db:
             if db("SELECT sid FROM session WHERE sid=%s", (new_sid,)):
                 raise TracError(_("Session '%(id)s' already exists. "
@@ -289,7 +332,7 @@ class Session(DetachedSession):
                       """, (sid,))
             elif len(authenticated_flags) == 1:
                 if not authenticated_flags[0]:
-                    # Update the anomymous session records so the session ID
+                    # Update the anonymous session records so the session ID
                     # becomes the user name, and set the authenticated flag.
                     self.env.log.debug("Promoting anonymous session %s to "
                                        "authenticated session for user %s",
@@ -315,7 +358,7 @@ class Session(DetachedSession):
         self._new = False
 
         self.sid = sid
-        self.bake_cookie(0) # expire the cookie
+        self.bake_cookie(0)  # expire the cookie
 
 
 class SessionAdmin(Component):
@@ -323,10 +366,12 @@ class SessionAdmin(Component):
 
     implements(IAdminCommandProvider)
 
+    request_handlers = ExtensionPoint(IRequestHandler)
+
     def get_admin_commands(self):
         hints = {
-           'datetime': get_datetime_format_hint(get_console_locale(self.env)),
-           'iso8601': get_datetime_format_hint('iso8601'),
+            'datetime': get_datetime_format_hint(get_console_locale(self.env)),
+            'iso8601': get_datetime_format_hint('iso8601'),
         }
         yield ('session list', '[sid[:0|1]] [...]',
                """List the name and email for the given sids
@@ -350,7 +395,8 @@ class SessionAdmin(Component):
                default if no suffix is specified).""",
                None, self._do_add)
 
-        yield ('session set', '<name|email> <sid[:0|1]> <value>',
+        yield ('session set', '<name|email|default_handler> '
+                              '<sid[:0|1]> <value>',
                """Set the name or email attribute of the given sid
 
                An sid suffix ':0' operates on an unauthenticated session with
@@ -375,13 +421,19 @@ class SessionAdmin(Component):
                format.""" % hints,
                None, self._do_purge)
 
+    @lazy
+    def _valid_default_handlers(self):
+        return sorted(handler.__class__.__name__
+                      for handler in self.request_handlers
+                      if is_valid_default_handler(handler))
+
     def _split_sid(self, sid):
         if sid.endswith(':0'):
-            return (sid[:-2], 0)
+            return sid[:-2], 0
         elif sid.endswith(':1'):
-            return (sid[:-2], 1)
+            return sid[:-2], 1
         else:
-            return (sid, 1)
+            return sid, 1
 
     def _get_sids(self):
         rows = self.env.db_query("SELECT sid, authenticated FROM session")
@@ -394,7 +446,7 @@ class SessionAdmin(Component):
                    if sid not in ('anonymous', 'authenticated', '*'))
         rows = self.env.db_query("""
             SELECT DISTINCT s.sid, s.authenticated, s.last_visit,
-                            n.value, e.value
+                            n.value, e.value, h.value
             FROM session AS s
               LEFT JOIN session_attribute AS n
                 ON (n.sid=s.sid AND n.authenticated=s.authenticated
@@ -402,12 +454,18 @@ class SessionAdmin(Component):
               LEFT JOIN session_attribute AS e
                 ON (e.sid=s.sid AND e.authenticated=s.authenticated
                     AND e.name='email')
+              LEFT JOIN session_attribute AS h
+                ON (h.sid=s.sid AND h.authenticated=s.authenticated
+                    AND h.name='default_handler')
             ORDER BY s.sid, s.authenticated
             """)
-        for sid, authenticated, last_visit, name, email in rows:
+        for sid, authenticated, last_visit, name, email, handler in rows:
             if all_anon and not authenticated or all_auth and authenticated \
                     or (sid, authenticated) in sids:
-                yield (sid, authenticated, last_visit, name, email)
+                yield (sid, authenticated,
+                       format_date(to_datetime(last_visit),
+                                   console_date_format),
+                       name, email, handler)
 
     def _complete_list(self, args):
         all_sids = self._get_sids() + ['*', 'anonymous', 'authenticated']
@@ -426,12 +484,9 @@ class SessionAdmin(Component):
     def _do_list(self, *sids):
         if not sids:
             sids = ['*']
-        print_table([(r[0], r[1], format_date(to_datetime(r[2]),
-                                              console_date_format),
-                      r[3], r[4])
-                     for r in self._get_list(sids)],
-                    [_('SID'), _('Auth'), _('Last Visit'), _('Name'),
-                     _('Email')])
+        headers = (_("SID"), _("Auth"), _("Last Visit"), _("Name"),
+                   _("Email"), _("Default Handler"))
+        print_table(self._get_list(sids), headers)
 
     def _do_add(self, sid, name=None, email=None):
         sid, authenticated = self._split_sid(sid)
@@ -442,18 +497,22 @@ class SessionAdmin(Component):
             except Exception:
                 raise AdminCommandError(_("Session '%(sid)s' already exists",
                                           sid=sid))
-            if name is not None:
+            if name:
                 db("INSERT INTO session_attribute VALUES (%s,%s,'name',%s)",
                     (sid, authenticated, name))
-            if email is not None:
+            if email:
                 db("INSERT INTO session_attribute VALUES (%s,%s,'email',%s)",
                     (sid, authenticated, email))
         self.env.invalidate_known_users_cache()
 
     def _do_set(self, attr, sid, val):
-        if attr not in ('name', 'email'):
+        if attr not in ('name', 'email', 'default_handler'):
             raise AdminCommandError(_("Invalid attribute '%(attr)s'",
                                       attr=attr))
+        if attr == 'default_handler':
+            if val and val not in self._valid_default_handlers:
+                raise AdminCommandError(_("Invalid default_handler '%(val)s'",
+                                          val=val))
         sid, authenticated = self._split_sid(sid)
         with self.env.db_transaction as db:
             if not db("""SELECT sid FROM session
@@ -465,8 +524,9 @@ class SessionAdmin(Component):
                 DELETE FROM session_attribute
                 WHERE sid=%s AND authenticated=%s AND name=%s
                 """, (sid, authenticated, attr))
-            db("INSERT INTO session_attribute VALUES (%s, %s, %s, %s)",
-               (sid, authenticated, attr, val))
+            if val:
+                db("INSERT INTO session_attribute VALUES (%s, %s, %s, %s)",
+                   (sid, authenticated, attr, val))
         self.env.invalidate_known_users_cache()
 
     def _do_delete(self, *sids):
@@ -499,6 +559,17 @@ class SessionAdmin(Component):
             db("""
                 DELETE FROM session_attribute
                 WHERE authenticated=0
-                      AND sid NOT IN (SELECT sid FROM session
-                                      WHERE authenticated=0)
+                      AND NOT EXISTS (SELECT * FROM session AS s
+                                      WHERE s.sid=session_attribute.sid
+                                      AND s.authenticated=0)
                 """)
+
+
+def get_session_attribute(env, sid, authenticated, name, default=None):
+    for row in env.db_query("""
+            SELECT value FROM session_attribute
+            WHERE sid=%s AND authenticated=%s AND name=%s
+            """, (sid, 1 if authenticated else 0, name)):
+        return row[0]
+    else:
+        return default
