@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007-2009 Edgewall Software
+# Copyright (C) 2007-2018 Edgewall Software
 # Copyright (C) 2007 Alec Thomas <alec@swapoff.org>
 # All rights reserved.
 #
@@ -14,21 +14,16 @@
 #
 # Author: Alec Thomas <alec@swapoff.org>
 
+import os
 from fnmatch import fnmatchcase
 from itertools import groupby
-import os
 
-from trac.core import *
-from trac.config import ConfigurationError, Option
-from trac.perm import PermissionSystem, IPermissionPolicy
-from trac.util import lazy
-from trac.util.text import to_unicode
-
-ConfigObj = None
-try:
-    from configobj import ConfigObj, ConfigObjError
-except ImportError:
-    pass
+from trac.config import ConfigurationError, ParsingError, PathOption, \
+                        UnicodeConfigParser
+from trac.core import Component, implements
+from trac.perm import IPermissionPolicy, PermissionSystem
+from trac.util import to_list
+from trac.util.text import exception_to_unicode
 
 
 class AuthzPolicy(Component):
@@ -38,25 +33,20 @@ class AuthzPolicy(Component):
     supported.
 
     As the fine-grained permissions brought by this permission policy are
-    often used in complement of the other pemission policies (like the
+    often used in complement of the other permission policies (like the
     `DefaultPermissionPolicy`), there's no need to redefine all the
     permissions here. Only additional rights or restrictions should be added.
 
     === Installation ===
-    Note that this plugin requires the `configobj` package:
+    Enabling this policy requires listing it in `trac.ini`::
 
-        http://www.voidspace.org.uk/python/configobj.html
+      {{{
+      [trac]
+      permission_policies = AuthzPolicy, DefaultPermissionPolicy
 
-    You should be able to install it by doing a simple `easy_install configobj`
-
-    Enabling this policy requires listing it in `trac.ini:
-    {{{
-    [trac]
-    permission_policies = AuthzPolicy, DefaultPermissionPolicy
-
-    [authz_policy]
-    authz_file = conf/authzpolicy.conf
-    }}}
+      [authz_policy]
+      authz_file = conf/authzpolicy.conf
+      }}}
 
     This means that the `AuthzPolicy` permissions will be checked first, and
     only if no rule is found will the `DefaultPermissionPolicy` be used.
@@ -66,31 +56,35 @@ class AuthzPolicy(Component):
     The `authzpolicy.conf` file is a `.ini` style configuration file.
 
      - Each section of the config is a glob pattern used to match against a
-       Trac resource descriptor. These descriptors are in the form:
-       {{{
-       <realm>:<id>@<version>[/<realm>:<id>@<version> ...]
-       }}}
+       Trac resource descriptor. These descriptors are in the form::
+
+         {{{
+         <realm>:<id>@<version>[/<realm>:<id>@<version> ...]
+         }}}
+
        Resources are ordered left to right, from parent to child. If any
        component is inapplicable, `*` is substituted. If the version pattern is
        not specified explicitely, all versions (`@*`) is added implicitly
 
-       Example: Match the WikiStart page
-       {{{
-       [wiki:*]
-       [wiki:WikiStart*]
-       [wiki:WikiStart@*]
-       [wiki:WikiStart]
-       }}}
+       Example: Match the WikiStart page::
 
-       Example: Match the attachment `wiki:WikiStart@117/attachment/FOO.JPG@*`
-       on WikiStart
-       {{{
-       [wiki:*]
-       [wiki:WikiStart*]
-       [wiki:WikiStart@*]
-       [wiki:WikiStart@*/attachment/*]
-       [wiki:WikiStart@117/attachment/FOO.JPG]
-       }}}
+         {{{
+         [wiki:*]
+         [wiki:WikiStart*]
+         [wiki:WikiStart@*]
+         [wiki:WikiStart]
+         }}}
+
+       Example: Match the attachment
+       ``wiki:WikiStart@117/attachment/FOO.JPG@*`` on WikiStart::
+
+         {{{
+         [wiki:*]
+         [wiki:WikiStart*]
+         [wiki:WikiStart@*]
+         [wiki:WikiStart@*/attachment/*]
+         [wiki:WikiStart@117/attachment/FOO.JPG]
+         }}}
 
      - Sections are checked against the current Trac resource '''IN ORDER''' of
        appearance in the configuration file. '''ORDER IS CRITICAL'''.
@@ -101,105 +95,102 @@ class AuthzPolicy(Component):
        denied rather than granted. The username will match any of 'anonymous',
        'authenticated', <username> or '*', using normal Trac permission rules.
 
-    Example configuration:
-    {{{
-    [groups]
-    administrators = athomas
+    Example configuration::
 
-    [*/attachment:*]
-    * = WIKI_VIEW, TICKET_VIEW
+      {{{
+      [groups]
+      administrators = athomas
 
-    [wiki:WikiStart@*]
-    @administrators = WIKI_ADMIN
-    anonymous = WIKI_VIEW
-    * = WIKI_VIEW
+      [*/attachment:*]
+      * = WIKI_VIEW, TICKET_VIEW
 
-    # Deny access to page templates
-    [wiki:PageTemplates/*]
-    * =
+      [wiki:WikiStart@*]
+      @administrators = WIKI_ADMIN
+      anonymous = WIKI_VIEW
+      * = WIKI_VIEW
+
+      # Deny access to page templates
+      [wiki:PageTemplates/*]
+      * =
+
+      # Match everything else
+      [*]
+      @administrators = TRAC_ADMIN
+      anonymous = BROWSER_VIEW, CHANGESET_VIEW, FILE_VIEW, LOG_VIEW,
+          MILESTONE_VIEW, POLL_VIEW, REPORT_SQL_VIEW, REPORT_VIEW,
+          ROADMAP_VIEW, SEARCH_VIEW, TICKET_CREATE, TICKET_MODIFY,
+          TICKET_VIEW, TIMELINE_VIEW,
+          WIKI_CREATE, WIKI_MODIFY, WIKI_VIEW
+      # Give authenticated users some extra permissions
+      authenticated = REPO_SEARCH, XML_RPC
+      }}}
 
-    # Match everything else
-    [*]
-    @administrators = TRAC_ADMIN
-    anonymous = BROWSER_VIEW, CHANGESET_VIEW, FILE_VIEW, LOG_VIEW,
-        MILESTONE_VIEW, POLL_VIEW, REPORT_SQL_VIEW, REPORT_VIEW, ROADMAP_VIEW,
-        SEARCH_VIEW, TICKET_CREATE, TICKET_MODIFY, TICKET_VIEW, TIMELINE_VIEW,
-        WIKI_CREATE, WIKI_MODIFY, WIKI_VIEW
-    # Give authenticated users some extra permissions
-    authenticated = REPO_SEARCH, XML_RPC
-    }}}
     """
     implements(IPermissionPolicy)
 
-    authz_file = Option('authz_policy', 'authz_file', '',
-                        'Location of authz policy configuration file.')
+    authz_file = PathOption('authz_policy', 'authz_file', '',
+                            "Location of authz policy configuration file. "
+                            "Non-absolute paths are relative to the "
+                            "Environment `conf` directory.")
 
-    authz = None
-    authz_mtime = None
+    def __init__(self):
+        self.authz = None
+        self.authz_mtime = None
+        self.groups_by_user = {}
 
     # IPermissionPolicy methods
 
     def check_permission(self, action, username, resource, perm):
         if not self.authz_mtime or \
-                os.path.getmtime(self.get_authz_file) > self.authz_mtime:
+                os.path.getmtime(self.authz_file) != self.authz_mtime:
             self.parse_authz()
         resource_key = self.normalise_resource(resource)
         self.log.debug('Checking %s on %s', action, resource_key)
         permissions = self.authz_permissions(resource_key, username)
         if permissions is None:
             return None                 # no match, can't decide
-        elif permissions == ['']:
+        elif permissions == []:
             return False                # all actions are denied
 
         # FIXME: expand all permissions once for all
         ps = PermissionSystem(self.env)
         for deny, perms in groupby(permissions,
                                    key=lambda p: p.startswith('!')):
-            if deny and action in ps.expand_actions([p[1:] for p in perms]):
+            if deny and action in ps.expand_actions(p[1:] for p in perms):
                 return False            # action is explicitly denied
             elif action in ps.expand_actions(perms):
-                return True            # action is explicitly granted
+                return True             # action is explicitly granted
 
-        return None                    # no match for action, can't decide
+        return None                     # no match for action, can't decide
 
     # Internal methods
 
-    @lazy
-    def get_authz_file(self):
-        if not self.authz_file:
-            self.log.error('The `[authz_policy] authz_file` configuration '
-                           'option in trac.ini is empty or not defined.')
-            raise ConfigurationError()
-
-        authz_file = self.authz_file if os.path.isabs(self.authz_file) \
-                                     else os.path.join(self.env.path,
-                                                       self.authz_file)
-        try:
-            os.stat(authz_file)
-        except OSError, e:
-            self.log.error("Error parsing authz permission policy file: %s",
-                           to_unicode(e))
-            raise ConfigurationError()
-        return authz_file
-
     def parse_authz(self):
-        if ConfigObj is None:
-            self.log.error("ConfigObj package not found.")
-            raise ConfigurationError()
         self.log.debug("Parsing authz security policy %s",
-                       self.get_authz_file)
+                       self.authz_file)
+
+        if not self.authz_file:
+            self.log.error("The `[authz_policy] authz_file` configuration "
+                           "option in trac.ini is empty or not defined.")
+            raise ConfigurationError()
         try:
-            self.authz = ConfigObj(self.get_authz_file, encoding='utf8',
-                                   raise_errors=True)
-        except ConfigObjError, e:
+            self.authz_mtime = os.path.getmtime(self.authz_file)
+        except OSError as e:
             self.log.error("Error parsing authz permission policy file: %s",
-                           to_unicode(e))
+                           exception_to_unicode(e))
+            raise ConfigurationError()
+
+        self.authz = UnicodeConfigParser(ignorecase_option=False)
+        try:
+            self.authz.read(self.authz_file)
+        except ParsingError as e:
+            self.log.error("Error parsing authz permission policy file: %s",
+                           exception_to_unicode(e))
             raise ConfigurationError()
         groups = {}
-        for group, users in self.authz.get('groups', {}).iteritems():
-            if isinstance(users, basestring):
-                users = [users]
-            groups[group] = map(to_unicode, users)
+        if self.authz.has_section('groups'):
+            for group, users in self.authz.items('groups'):
+                groups[group] = to_list(users)
 
         self.groups_by_user = {}
 
@@ -213,7 +204,19 @@ class AuthzPolicy(Component):
         for group, users in groups.iteritems():
             add_items('@' + group, users)
 
-        self.authz_mtime = os.path.getmtime(self.get_authz_file)
+        all_actions = set(PermissionSystem(self.env).get_actions())
+        authz_basename = os.path.basename(self.authz_file)
+        for section in self.authz.sections():
+            if section == 'groups':
+                continue
+            for _, actions in self.authz.items(section):
+                for action in to_list(actions):
+                    if action.startswith('!'):
+                        action = action[1:]
+                    if action not in all_actions:
+                        self.log.warning("The action %s in the [%s] section "
+                                         "of %s is not a valid action.",
+                                         action, section, authz_basename)
 
     def normalise_resource(self, resource):
         def to_descriptor(resource):
@@ -245,22 +248,21 @@ class AuthzPolicy(Component):
         # TODO: Handle permission negation in sections. eg. "if in this
         # ticket, remove TICKET_MODIFY"
         if username and username != 'anonymous':
-            valid_users = ['*', 'authenticated', username]
+            valid_users = ['*', 'authenticated', 'anonymous', username]
         else:
             valid_users = ['*', 'anonymous']
-        for resource_section in [a for a in self.authz.sections
+        for resource_section in [a for a in self.authz.sections()
                                    if a != 'groups']:
-            resource_glob = to_unicode(resource_section)
+            resource_glob = resource_section
             if '@' not in resource_glob:
                 resource_glob += '@*'
 
             if fnmatchcase(resource_key, resource_glob):
-                section = self.authz[resource_section]
-                for who, permissions in section.iteritems():
-                    who = to_unicode(who)
+                for who, permissions in self.authz.items(resource_section):
+                    permissions = to_list(permissions)
                     if who in valid_users or \
                             who in self.groups_by_user.get(username, []):
-                        self.log.debug('%s matched section %s for user %s',
+                        self.log.debug("%s matched section %s for user %s",
                                        resource_key, resource_glob, username)
                         if isinstance(permissions, basestring):
                             return [permissions]
