@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2011 Edgewall Software
+# Copyright (C) 2005-2018 Edgewall Software
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005-2007 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
@@ -48,30 +48,39 @@ Warning:
 """
 
 import os.path
+import re
 import weakref
 import posixpath
+from urllib import quote
 
-from trac.config import ListOption
+from trac.config import ListOption, ChoiceOption
 from trac.core import *
 from trac.env import ISystemInfoProvider
 from trac.versioncontrol import Changeset, Node, Repository, \
-                                IRepositoryConnector, \
+                                IRepositoryConnector, InvalidRepository, \
                                 NoSuchChangeset, NoSuchNode
 from trac.versioncontrol.cache import CachedRepository
 from trac.util import embedded_numbers
-from trac.util.text import exception_to_unicode, to_unicode
+from trac.util.concurrency import threading
+from trac.util.text import exception_to_unicode, to_unicode, to_utf8
 from trac.util.translation import _
-from trac.util.datefmt import from_utimestamp
+from trac.util.datefmt import from_utimestamp, to_datetime, utc
+from trac.web.href import Href
 
 
 application_pool = None
+application_pool_lock = threading.Lock()
 
 
 def _import_svn():
-    global fs, repos, core, delta, _kindmap
+    global fs, repos, core, delta, _kindmap, _svn_uri_canonicalize
     from svn import fs, repos, core, delta
     _kindmap = {core.svn_node_dir: Node.DIRECTORY,
                 core.svn_node_file: Node.FILE}
+    try:
+        _svn_uri_canonicalize = core.svn_uri_canonicalize  # Subversion 1.7+
+    except AttributeError:
+        _svn_uri_canonicalize = lambda v: v
     # Protect svn.core methods from GC
     Pool.apr_pool_clear = staticmethod(core.apr_pool_clear)
     Pool.apr_pool_destroy = staticmethod(core.apr_pool_destroy)
@@ -150,19 +159,21 @@ class Pool(object):
         """Create a new memory pool"""
 
         global application_pool
-        self._parent_pool = parent_pool or application_pool
 
-        # Create pool
-        if self._parent_pool:
-            self._pool = core.svn_pool_create(self._parent_pool())
-        else:
-            # If we are an application-level pool,
-            # then initialize APR and set this pool
-            # to be the application-level pool
-            core.apr_initialize()
-            application_pool = self
+        with application_pool_lock:
+            self._parent_pool = parent_pool or application_pool
+
+            # Create pool
+            if self._parent_pool:
+                self._pool = core.svn_pool_create(self._parent_pool())
+            else:
+                # If we are an application-level pool,
+                # then initialize APR and set this pool
+                # to be the application-level pool
+                core.apr_initialize()
+                self._pool = core.svn_pool_create(None)
+                application_pool = self
 
-            self._pool = core.svn_pool_create(None)
         self._mark_valid()
 
     def __call__(self):
@@ -245,6 +256,9 @@ class SvnCachedRepository(CachedRepository):
     def rev_db(self, rev):
         return int(rev or 0)
 
+    def normalize_rev(self, rev):
+        return self.repos.normalize_rev(rev)
+
 
 class SubversionConnector(Component):
 
@@ -265,17 +279,28 @@ class SubversionConnector(Component):
         Example: `/tags/*, /projectAlpha/tags/A-1.0, /projectAlpha/tags/A-v1.1`
         """)
 
+    eol_style = ChoiceOption(
+        'svn', 'eol_style', ['native', 'LF', 'CRLF', 'CR'], doc=
+        """End-of-Line character sequences when `svn:eol-style` property is
+        `native`.
+
+        If `native`, substitute with the native EOL marker on the server.
+        Otherwise, if `LF`, `CRLF` or `CR`, substitute with the specified
+        EOL marker.
+
+        (''since 1.0.2'')""")
+
     error = None
 
     def __init__(self):
         self._version = None
         try:
             _import_svn()
-            self.log.debug('Subversion bindings imported')
-        except ImportError, e:
+        except ImportError as e:
             self.error = e
             self.log.info('Failed to load Subversion bindings', exc_info=True)
         else:
+            self.log.debug("Subversion bindings imported")
             version = (core.SVN_VER_MAJOR, core.SVN_VER_MINOR,
                        core.SVN_VER_MICRO)
             self._version = '%d.%d.%d' % version + core.SVN_VER_TAG
@@ -287,7 +312,7 @@ class SubversionConnector(Component):
     # ISystemInfoProvider methods
 
     def get_system_info(self):
-        if self._version is not None:
+        if self._version:
             yield 'Subversion', self._version
 
     # IRepositoryConnector methods
@@ -307,6 +332,7 @@ class SubversionConnector(Component):
         'direct-svnfs'.
         """
         params.update(tags=self.tags, branches=self.branches)
+        params.setdefault('eol_style', self.eol_style)
         repos = SubversionRepository(dir, params, self.log)
         if type != 'direct-svnfs':
             repos = SvnCachedRepository(self.env, repos, self.log)
@@ -328,20 +354,23 @@ class SubversionRepository(Repository):
         else: # note that this should usually not happen (unicode arg expected)
             path_utf8 = to_unicode(path).encode('utf-8')
 
-        path_utf8 = os.path.normpath(path_utf8).replace('\\', '/')
+        path_utf8 = core.svn_path_canonicalize(
+                                os.path.normpath(path_utf8).replace('\\', '/'))
         self.path = path_utf8.decode('utf-8')
 
         root_path_utf8 = repos.svn_repos_find_root_path(path_utf8, self.pool())
         if root_path_utf8 is None:
-            raise TracError(_("%(path)s does not appear to be a Subversion "
-                              "repository.", path=to_unicode(path_utf8)))
+            raise InvalidRepository(
+                _("%(path)s does not appear to be a Subversion repository.",
+                  path=to_unicode(path_utf8)))
 
         try:
             self.repos = repos.svn_repos_open(root_path_utf8, self.pool())
-        except core.SubversionException, e:
-            raise TracError(_("Couldn't open Subversion repository %(path)s: "
-                              "%(svn_error)s", path=to_unicode(path_utf8),
-                              svn_error=exception_to_unicode(e)))
+        except core.SubversionException as e:
+            raise InvalidRepository(
+                _("Couldn't open Subversion repository %(path)s: "
+                  "%(svn_error)s", path=to_unicode(path_utf8),
+                  svn_error=exception_to_unicode(e)))
         self.fs_ptr = repos.svn_repos_fs(self.repos)
 
         self.uuid = fs.get_uuid(self.fs_ptr, self.pool())
@@ -361,7 +390,8 @@ class SubversionRepository(Repository):
         assert self.scope[0] == '/'
         # we keep root_path_utf8 for  RA
         ra_prefix = 'file:///' if os.name == 'nt' else 'file://'
-        self.ra_url_utf8 = ra_prefix + root_path_utf8
+        self.ra_url_utf8 = _svn_uri_canonicalize(ra_prefix +
+                                                 quote(root_path_utf8))
         self.clear()
 
     def clear(self, youngest_rev=None):
@@ -399,11 +429,17 @@ class SubversionRepository(Repository):
             return self.youngest_rev
         else:
             try:
-                rev = int(rev)
-                if rev <= self.youngest_rev:
-                    return rev
-            except (ValueError, TypeError):
-                pass
+                normrev = int(rev)
+                if 0 <= normrev <= self.youngest_rev:
+                    return normrev
+                else:
+                    self.log.debug("%r cannot be normalized in %s: out of [0, "
+                                   "%r]", rev, self.reponame or '(default)',
+                                   self.youngest_rev)
+            except (ValueError, TypeError) as e:
+                self.log.debug("%r cannot be normalized in %s: %s", rev,
+                               self.reponame or '(default)',
+                               exception_to_unicode(e))
             raise NoSuchChangeset(rev)
 
     def close(self):
@@ -456,9 +492,10 @@ class SubversionRepository(Repository):
         """
         url = self.params.get('url', '').rstrip('/')
         if url:
-            if not path or path == '/':
-                return url
-            return url + '/' + path.lstrip('/')
+            href = Href(url)
+            if path:
+                path = path.lstrip('/')
+            return href(path)
 
     def get_changeset(self, rev):
         """Produce a `SubversionChangeset` from given revision
@@ -475,7 +512,7 @@ class SubversionRepository(Repository):
         specifications. No revision given means use the latest.
         """
         path = path or ''
-        if path and path[-1] == '/':
+        if path and path != '/' and path[-1] == '/':
             path = path[:-1]
         rev = self.normalize_rev(rev) or self.youngest_rev
         return SubversionNode(path, rev, self, self.pool)
@@ -493,6 +530,18 @@ class SubversionRepository(Repository):
             revs.append(r)
         return revs
 
+    def _get_changed_revs(self, node_infos):
+        path_revs = {}
+        for node, first in node_infos:
+            path = node.path
+            revs = []
+            for p, r, chg in node.get_history():
+                if p != path or r < first:
+                    break
+                revs.append(r)
+            path_revs[path] = revs
+        return path_revs
+
     def _history(self, path, start, end, pool):
         """`path` is a unicode path in the scope.
 
@@ -640,14 +689,6 @@ class SubversionRepository(Repository):
 
         (wraps ``repos.svn_repos_dir_delta``)
         """
-        def key(value):
-            return value[1].path if value[1] is not None else value[0].path
-        return iter(sorted(self._get_changes(old_path, old_rev, new_path,
-                                             new_rev, ignore_ancestry),
-                           key=key))
-
-    def _get_changes(self, old_path, old_rev, new_path, new_rev,
-                     ignore_ancestry):
         old_node = new_node = None
         old_rev = self.normalize_rev(old_rev)
         new_rev = self.normalize_rev(new_rev)
@@ -688,8 +729,12 @@ class SubversionRepository(Repository):
                                       entry_props,
                                       ignore_ancestry,
                                       subpool())
-            for path, kind, change in editor.deltas:
-                path = _from_svn(path)
+            # sort deltas by path before creating `SubversionNode`s to reduce
+            # memory usage (#10978)
+            deltas = sorted(((_from_svn(path), kind, change)
+                             for path, kind, change in editor.deltas),
+                            key=lambda entry: entry[0])
+            for path, kind, change in deltas:
                 old_node = new_node = None
                 if change != Changeset.ADD:
                     old_node = self.get_node(posixpath.join(old_path, path),
@@ -727,7 +772,10 @@ class SubversionNode(Node):
         if parent_root:
             self.root = parent_root
         else:
-            self.root = fs.revision_root(self.fs_ptr, rev, pool)
+            try:
+                self.root = fs.revision_root(self.fs_ptr, rev, pool)
+            except core.SubversionException as e:
+                raise NoSuchNode(path, rev, exception_to_unicode(e))
         node_type = fs.check_path(self.root, self._scoped_path_utf8, pool)
         if not node_type in _kindmap:
             raise NoSuchNode(path, rev)
@@ -753,13 +801,15 @@ class SubversionNode(Node):
         """Retrieve raw content as a "read()"able object."""
         if self.isdir:
             return None
-        pool = Pool(self.pool)
-        s = core.Stream(fs.file_contents(self.root, self._scoped_path_utf8,
-                                         pool()))
-        # The stream object needs to reference the pool to make sure the pool
-        # is not destroyed before the former.
-        s._pool = pool
-        return s
+        return FileContentStream(self)
+
+    def get_processed_content(self, keyword_substitution=True, eol_hint=None):
+        """Retrieve processed content as a "read()"able object."""
+        if self.isdir:
+            return None
+        eol_style = self.repos.params.get('eol_style') if eol_hint is None \
+            else eol_hint
+        return FileContentStream(self, keyword_substitution, eol_style)
 
     def get_entries(self):
         """Yield `SubversionNode` corresponding to entries in this directory.
@@ -811,13 +861,16 @@ class SubversionNode(Node):
                 rev = _svn_rev(self.rev)
                 start = _svn_rev(0)
                 file_url_utf8 = posixpath.join(self.repos.ra_url_utf8,
-                                               self._scoped_path_utf8)
+                                               quote(self._scoped_path_utf8))
+                # svn_client_blame2() requires a canonical uri since
+                # Subversion 1.7 (#11167)
+                file_url_utf8 = _svn_uri_canonicalize(file_url_utf8)
                 self.repos.log.info('opening ra_local session to %r',
                                     file_url_utf8)
                 from svn import client
                 client.blame2(file_url_utf8, rev, start, rev, blame_receiver,
                               client.create_context(), self.pool())
-            except (core.SubversionException, AttributeError), e:
+            except (core.SubversionException, AttributeError) as e:
                 # svn thinks file is a binary or blame not supported
                 raise TracError(_('svn blame failed on %(path)s: %(error)s',
                                   path=self.path, error=to_unicode(e)))
@@ -918,6 +971,7 @@ class SubversionNode(Node):
 class SubversionChangeset(Changeset):
 
     def __init__(self, repos, rev, scope, pool=None):
+        self.log = repos.log
         self.rev = rev
         self.scope = scope
         self.fs_ptr = repos.fs_ptr
@@ -1006,7 +1060,7 @@ class SubversionChangeset(Changeset):
                 action = Changeset.EDIT
                 # identify the most interesting base_path/base_rev
                 # in terms of last changed information (see r2562)
-                if revroots.has_key(base_rev):
+                if base_rev in revroots:
                     b_root = revroots[base_rev]
                 else:
                     b_root = fs.revision_root(self.fs_ptr, base_rev, pool())
@@ -1044,7 +1098,11 @@ class SubversionChangeset(Changeset):
             yield tuple(change)
 
     def _get_prop(self, name):
-        return fs.revision_prop(self.fs_ptr, self.rev, name, self.pool())
+        try:
+            return fs.revision_prop(self.fs_ptr, self.rev, name, self.pool())
+        except core.SubversionException as e:
+            self.log.debug("%r of the %r cannot be retrieved", name, self.rev)
+            raise
 
 
 #
@@ -1094,3 +1152,226 @@ def DiffChangeEditor():
 
     return DiffChangeEditor()
 
+
+class FileContentStream(object):
+
+    KEYWORD_GROUPS = {
+        'rev': ['LastChangedRevision', 'Rev', 'Revision'],
+        'date': ['LastChangedDate', 'Date'],
+        'author': ['LastChangedBy', 'Author'],
+        'url': ['HeadURL', 'URL'],
+        'id': ['Id'],
+        'header': ['Header'],
+        }
+    KEYWORDS = reduce(set.union, map(set, KEYWORD_GROUPS.values()))
+    KEYWORD_SPLIT_RE = re.compile(r'[ \t\v\n\b\r\f]+')
+    KEYWORD_EXPAND_RE = re.compile(r'%[abdDPrRu_%HI]')
+    NATIVE_EOL = '\r\n' if os.name == 'nt' else '\n'
+    NEWLINES = {'LF': '\n', 'CRLF': '\r\n', 'CR': '\r', 'native': NATIVE_EOL}
+    KEYWORD_MAX_SIZE = 255
+    CHUNK_SIZE = 4096
+
+    keywords_re = None
+    native_eol = None
+    newline = '\n'
+
+    def __init__(self, node, keyword_substitution=None, eol=None):
+        self.translated = ''
+        self.buffer = ''
+        self.repos = node.repos
+        self.node = node
+        self.fs_ptr = node.fs_ptr
+        self.pool = Pool()
+        # Note: we _must_ use a detached pool here, as the lifetime of
+        # this object can exceed those of the node or even the repository
+        if keyword_substitution:
+            self.keywords = self._get_keyword_values(
+                                        node._get_prop(core.SVN_PROP_KEYWORDS))
+            self.keywords_re = self._build_keywords_re(self.keywords)
+        if self.NEWLINES.get(eol, '\n') != '\n' and \
+           node._get_prop(core.SVN_PROP_EOL_STYLE) == 'native':
+            self.native_eol = True
+            self.newline = self.NEWLINES[eol]
+        self.stream = core.Stream(fs.file_contents(node.root,
+                                                   node._scoped_path_utf8,
+                                                   self.pool()))
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, et, ev, tb):
+        self.close()
+
+    def __del__(self):
+        self.close()
+
+    def close(self):
+        self.stream = None
+        self.fs_ptr = None
+        if self.pool:
+            self.pool.destroy()
+            self.pool = None
+
+    def read(self, n=None):
+        if self.stream is None:
+            raise ValueError('I/O operation on closed file')
+        if self.keywords_re is None and not self.native_eol:
+            return self._read_dumb(self.stream, n)
+        else:
+            return self._read_substitute(self.stream, n)
+
+    def _get_revprop(self, name, rev):
+        return fs.revision_prop(self.fs_ptr, rev, name, self.pool())
+
+    def _split_keywords(self, keywords):
+        return filter(None, self.KEYWORD_SPLIT_RE.split(keywords or ''))
+
+    def _get_keyword_values(self, keywords):
+        keywords = self._split_keywords(keywords)
+        if not keywords:
+            return None
+
+        node = self.node
+        mtime = to_datetime(node.last_modified, utc)
+        shortdate = self._format_shortdate(mtime)
+        longdate = self._format_longdate(mtime)
+        created_rev = unicode(node.created_rev)
+        # Note that the `to_unicode` has a small probability to mess-up binary
+        # properties, see #4321.
+        author = to_unicode(self._get_revprop(core.SVN_PROP_REVISION_AUTHOR,
+                                              node.created_rev))
+        path = node.path.lstrip('/')
+        url = node.repos.get_path_url(path, node.rev) or path
+        root_url = node.repos.get_path_url('', node.rev) or '/'
+        id_ = ' '.join((node.name, created_rev, shortdate, author))
+        data = {
+            'rev': created_rev, 'author': author, 'url': url, 'date': longdate,
+            'id': id_,
+            'header': ' '.join((url, created_rev, shortdate, author)),
+            '%a': author, '%b': node.name, '%d': shortdate, '%D': longdate,
+            '%P': path, '%r': created_rev, '%R': root_url, '%u': url,
+            '%_': ' ', '%%': '%', '%I': id_,
+            '%H': ' '.join((path, created_rev, shortdate, author)),
+        }
+
+        def expand(match):
+            match = match.group(0)
+            return data.get(match, match)
+
+        values = {}
+        for name, aliases in self.KEYWORD_GROUPS.iteritems():
+            if any(kw in keywords for kw in aliases):
+                values.update((kw, data[name]) for kw in aliases)
+        for keyword in keywords:
+            if '=' not in keyword:
+                continue
+            name, definition = keyword.split('=', 1)
+            if name not in self.KEYWORDS:
+                values[name] = self.KEYWORD_EXPAND_RE.sub(expand, definition)
+
+        if values:
+            return dict((key, to_utf8(value))
+                        for key, value in values.iteritems())
+        else:
+            return None
+
+    def _build_keywords_re(self, keywords):
+        if keywords:
+            return re.compile("""
+                [$]
+                (?P<keyword>%s)
+                (?:
+                    :[ ][^$\r\n]+?[ ]   |
+                    ::[ ](?P<fixed>[^$\r\n]+?)[ #]
+                )?
+                [$]""" % '|'.join(map(re.escape, keywords)),
+                re.VERBOSE)
+        else:
+            return None
+
+    def _format_shortdate(self, mtime):
+        return mtime.strftime('%Y-%m-%d %H:%M:%SZ')
+
+    def _format_longdate(self, mtime):
+        text = mtime.strftime('%Y-%m-%d %H:%M:%S +0000 (%%(a)s, %d %%(b)s %Y)')
+        weekdays = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
+        months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
+                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
+        return text % {'a': weekdays[mtime.weekday()],
+                       'b': months[mtime.month - 1]}
+
+    def _read_dumb(self, stream, n):
+        return stream.read(n)
+
+    def _read_substitute(self, stream, n):
+        if n is None:
+            n = -1
+
+        buffer = self.buffer
+        translated = self.translated
+        while True:
+            if 0 <= n <= len(translated):
+                self.buffer = buffer
+                self.translated = translated[n:]
+                return translated[:n]
+
+            if len(buffer) < self.KEYWORD_MAX_SIZE:
+                buffer += stream.read(self.CHUNK_SIZE) or ''
+                if not buffer:
+                    self.buffer = buffer
+                    self.translated = ''
+                    return translated
+
+            # search first "$" character
+            pos = buffer.find('$') if self.keywords_re else -1
+            if pos == -1:
+                translated += self._translate_newline(buffer)
+                buffer = ''
+                continue
+            if pos > 0:
+                # move to the first "$" character
+                translated += self._translate_newline(buffer[:pos])
+                buffer = buffer[pos:]
+
+            match = None
+            while True:
+                # search second "$" character
+                pos = buffer.find('$', 1)
+                if pos == -1:
+                    translated += self._translate_newline(buffer)
+                    buffer = ''
+                    break
+                if pos < self.KEYWORD_MAX_SIZE:
+                    match = self.keywords_re.match(buffer)
+                    if match:
+                        break  # found "$Keyword$" in the first 255 bytes
+                # move to the second "$" character
+                translated += self._translate_newline(buffer[:pos])
+                buffer = buffer[pos:]
+            if pos == -1 or not match:
+                continue
+
+            # move to the next character of the second "$" character
+            pos += 1
+            translated += self._translate_keyword(buffer[:pos], match)
+            buffer = buffer[pos:]
+            continue
+
+    def _translate_newline(self, data):
+        if self.native_eol:
+            data = data.replace('\n', self.newline)
+        return data
+
+    def _translate_keyword(self, text, match):
+        keyword = match.group('keyword')
+        value = self.keywords.get(keyword)
+        if value is None:
+            return text
+        fixed = match.group('fixed')
+        if fixed is None:
+            n = self.KEYWORD_MAX_SIZE - len(keyword) - 5
+            return '$%s: %.*s $' % (keyword, n, value) if n >= 0 else text
+        else:
+            n = len(fixed)
+            return '$%s:: %-*.*s%s$' % \
+                   (keyword, n, n, value, '#' if n < len(value) else ' ')
